package mpi.eudico.client.annotator.interlinear.edit;

import java.awt.Color;
import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.List;

import mpi.eudico.client.annotator.interlinear.IGTTierType;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTAnnotation;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTDataModel;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionModel;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionViewerModel;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTTier;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTNodeRenderInfo;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTViewerRenderInfo;

/**
 * A renderer for suggestion sets. The rendering is similar to that of igt groups,
 * but adds numbering of the sets and horizontal or vertical orientation of the display.  
 * 
 * @author Han Sloetjes
 */
public class IGTSuggestionRenderer {
	
	private boolean debug = false;

	/**
	 * Renders the all suggestion sets in one row, in a horizontal layout.
	 * 
	 * @param g2d the graphics context
	 * @param model the suggestion viewer model
	 * @param vrInfo the top level viewer rendering information
	 * @param xOff the x offset in the graphics context
	 * @param yOff the y offset in the graphics context
	 * @param width the width for the entire suggestion model
	 * @param height the height for the suggestion model
	 */
	public void renderHorizontal(Graphics2D g2d, IGTSuggestionViewerModel svModel, IGTViewerRenderInfo vrInfo,
			int xOff, int yOff, int width, int height) {
		//g2d.setColor(Color.WHITE);
		//g2d.fillRect(xOff, yOff, width, height);
		int numRows = svModel.getRowCount();
		// render the suggestions header
		int startX = vrInfo.headerWidth;
		IGTSuggestionModel sugModel;
		
		for (int i = 0; i < numRows; i++) {
			sugModel = svModel.getRowData(i);
			g2d.setColor(svModel.getRenderInfo().headerBackGround);
			g2d.fillRect(startX, yOff, sugModel.getRenderInfo().width, svModel.getRenderInfo().headerHeight);
			g2d.setColor(Color.BLACK);
			g2d.drawRect(startX, yOff, sugModel.getRenderInfo().width, svModel.getRenderInfo().headerHeight);
			g2d.drawString("S" + (i + 1), startX + 2, svModel.getRenderInfo().headerHeight - 2);
			
			// draw block backgrounds
			g2d.setColor(svModel.getRenderInfo().blockBackGround);
			g2d.fillRect(startX, svModel.getRenderInfo().headerHeight + 1, sugModel.getRenderInfo().width, 
				sugModel.getRenderInfo().height);
			g2d.setColor(Color.LIGHT_GRAY);
			g2d.drawRect(startX, svModel.getRenderInfo().headerHeight + 1, sugModel.getRenderInfo().width, 
					sugModel.getRenderInfo().height);

			if (i == 0) {
				g2d.setColor(vrInfo.foregroundColor);
				g2d.drawLine(vrInfo.headerWidth, yOff, vrInfo.headerWidth, yOff + height);
			}
			startX += sugModel.getRenderInfo().width;
			startX += svModel.getRenderInfo().suggestionMargin;
		}
		
		// start rendering tiers, tier labels, annotations
		int tempY = yOff + svModel.getRenderInfo().headerHeight;
		int tempX = xOff;

		for (int i = 0; i < numRows; i++) {
			sugModel = svModel.getRowData(i);
			if (i == 0) {
				renderRowData(g2d, sugModel, vrInfo, true, tempX, tempY, 
						sugModel.getRenderInfo().width + vrInfo.headerWidth,
						sugModel.getRenderInfo().height);				
			} else {
				renderRowData(g2d, sugModel, vrInfo, false, tempX, tempY, 
						sugModel.getRenderInfo().width,
						sugModel.getRenderInfo().height);
			}

			
			tempX += sugModel.getRenderInfo().width;
			tempX += svModel.getRenderInfo().suggestionMargin;
			if (i == 0) {
				tempX += vrInfo.headerWidth;
			}
		}
		
	}
	
	/**
	 * Renders the annotations of a group of suggestions.
	 * 
	 * @param g2d the graphics context
	 * @param model the suggestion data model
	 * @param vrInfo the viewer model rendering information
	 * @param includeTierLabels if true the tier labels will be painted as well
	 * @param xOff the start x offset in the graphics context
	 * @param yOff the start y offset in the graphics context
	 * @param width the available width for the group of suggestions
	 * @param height the available height for the group of suggestions
	 */
	private void renderRowData(Graphics2D g2d, IGTDataModel model, IGTViewerRenderInfo vrInfo, boolean includeTierLabels,
			int xOff, int yOff, int width, int height) {
		List<IGTTier> renderedTier = new ArrayList<IGTTier>();
		IGTTier tier;
		IGTAnnotation igtAnn;
		IGTNodeRenderInfo annRender;
		int tempY = yOff;
		g2d.setColor(vrInfo.foregroundColor);
		
		for (int i = 0; i < model.getRowCount(); i++) {
			tier = model.getRowData(i);
			if (tier.getType() == IGTTierType.ROOT) {// skip the root, just a place holder
				continue;
			}
			if (renderedTier.contains(tier)) {
				continue;
			}
			int tempX = xOff;
			
			if (tier.isInWordLevelBlock()) {
				// tier labels
				if (includeTierLabels) {
					g2d.setColor(vrInfo.getColorForTier(tier));
					g2d.setFont(vrInfo.defaultFont);
					if (tier.getRenderInfo().getYPositions() != null) {
						for (int j = 0; j < tier.getRenderInfo().getYPositions().size(); j++) {
							// TODO position the baseline of the text correctly
							tempY = yOff + tier.getRenderInfo().getYPositions().get(j) + tier.getRenderInfo().height;
							g2d.drawString(model.getShortTierNameForIndex(i), tempX, tempY - 2);
						}
					} else {
						tempY = yOff + tier.getRenderInfo().y + tier.getRenderInfo().height;
						g2d.drawString(model.getShortTierNameForIndex(i), tempX, tempY - 2);
					}
					tempX += vrInfo.headerWidth;
				}
				
				// annotations
				g2d.setColor(vrInfo.foregroundColor);
				g2d.setFont(vrInfo.getFontForTier(tier.getTierName()));
				for (int j = 0; j < tier.getAnnotations().size(); j++) {
					igtAnn = tier.getAnnotations().get(j);
					annRender = igtAnn.getRenderInfo();
					tempY = yOff + annRender.y + annRender.height;
					
					g2d.drawString(igtAnn.getTextValue(), tempX + annRender.x, 
							tempY - 2);// TODO position the text correctly vertically
					if (debug) {
						g2d.setColor(Color.RED);
						g2d.drawRect(tempX + igtAnn.getRenderInfo().x, yOff + annRender.y, 
								igtAnn.getRenderInfo().calcWidth, annRender.height);
						g2d.setColor(vrInfo.foregroundColor);
					}
				}
			} else {
				// other types of tiers
				tempY = yOff + tier.getRenderInfo().y + tier.getRenderInfo().height;
				if (includeTierLabels) {
					g2d.setColor(vrInfo.getColorForTier(tier));
					g2d.setFont(vrInfo.defaultFont);
					g2d.drawString(model.getShortTierNameForIndex(i), tempX, tempY - 2); // position the baseline of the text correctly
					tempX += vrInfo.headerWidth;
				}
				
				// annotations
				g2d.setColor(vrInfo.foregroundColor);
				g2d.setFont(vrInfo.getFontForTier(tier.getTierName()));
				if (tier.getRenderInfo().numLines > 1) {
					if (tier.getAnnotations().size() >= 1) {
						igtAnn = tier.getAnnotations().get(0);
						List<String> lines = igtAnn.getRenderInfo().getWrappedLines();
						
						if (lines != null) {
							for (int a = 0; a < lines.size(); a++) {
								g2d.drawString(lines.get(a), tempX + igtAnn.getRenderInfo().x, tempY - 2);
								
								if (debug) {
									g2d.setColor(Color.RED);
									g2d.drawRect(tempX + igtAnn.getRenderInfo().x, tempY - tier.getRenderInfo().height, 
											igtAnn.getRenderInfo().calcWidth, tier.getRenderInfo().height);
									g2d.setColor(vrInfo.foregroundColor);
								}
								
								if (a != lines.size() - 1) {
									tempY += vrInfo.vertLineMargin;
									tempY += tier.getRenderInfo().height;
								}
							}
						} 				
					} 
				} else {
					if (tier.getAnnotations().size() >= 1) {
						igtAnn = tier.getAnnotations().get(0);
						g2d.drawString(igtAnn.getTextValue(), tempX + igtAnn.getRenderInfo().x, tempY - 2);
						if (debug) {
							g2d.setColor(Color.RED);
							g2d.drawRect(tempX + igtAnn.getRenderInfo().x, tempY - tier.getRenderInfo().height, 
									igtAnn.getRenderInfo().calcWidth, tier.getRenderInfo().height);
							g2d.setColor(vrInfo.foregroundColor);
						}
					}
				}
			}
			
			renderedTier.add(tier);
			tempY += vrInfo.vertLineMargin;//??
		}
		
	}
	
	public void renderRowVertical(Graphics2D g2d, IGTDataModel model, IGTViewerRenderInfo vrInfo, int row,
			int xOff, int yOff, int width, int height) {
		// TODO implement
	}
	
	public void renderGrid(Graphics2D g2d, IGTDataModel model, IGTViewerRenderInfo vrInfo,
			int xOff, int yOff, int width, int height) {
		// TODO implement, if possible/necessary
	}
}
