package mpi.eudico.client.annotator.interlinear.edit;

import java.util.ArrayList;
import java.util.List;

import mpi.eudico.client.annotator.interlinear.edit.event.IGTDataModelEvent;
import mpi.eudico.client.annotator.interlinear.edit.event.IGTDataModelListener;
import mpi.eudico.client.annotator.interlinear.edit.event.IGTViewerModelEvent;
import mpi.eudico.client.annotator.interlinear.edit.event.IGTViewerModelListener;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTDataModel;

/**
 * Implementation of a viewer model for interlinear blocks in the form of 
 * IGTDataModel objects.
 * 
 * @author Han Sloetjes
 */
public class IGTViewerModelImpl implements IGTViewerModel, IGTDataModelListener {
	private List<IGTDataModel> rowData;
	private int editingRow = -1;
	private boolean isExpanded = false;
	private List<IGTViewerModelListener> listeners;
	
	private List<String> visibleTiers;
	private List<String> visibleParticipants;
	private List<String> visibleTiersCollapsed;
	
	/**
	 * No-arg constructor.
	 */
	public IGTViewerModelImpl() {
		rowData = new ArrayList<IGTDataModel>();
		listeners = new ArrayList<IGTViewerModelListener>(4);
	}

	@Override
	public int getRowCount() {
		return rowData.size();
	}

	@Override
	public int getEditingRow() {
		return editingRow;
	}
	
	@Override
	public void startEditingRow(int row) {
		editingRow = row;
	}

	@Override
	public boolean getExpandedState() {
		return isExpanded;
	}

	@Override
	public void setExpandedState(boolean expanded) {
		isExpanded = expanded;
	}
	
	@Override
	public void setVisibleTiersInCollapsedState(List<String> tierNames) {
		// TODO copy the names to the list instead?
		visibleTiersCollapsed = tierNames;

	}

	@Override
	public List<String> getVisibleTiersInCollapsedState() {
		// TODO don't return a reference to the list itself, but copy values?
		return visibleTiersCollapsed;
	}

	@Override
	public void setVisibleTiers(List<String> tierNames) {
		// TODO copy the values to the local list instead?
		visibleTiers = tierNames;
	}
	
	@Override
	public List<String> getVisibleTiers() {
		// TODO return a copy of the list instead of a reference?
		return visibleTiers;
	}

	@Override
	public void setVisibleParticipants(List<String> partNames) {
		// TODO copy the values instead of setting the reference?
		this.visibleParticipants = partNames;
	}
	
	@Override
	public List<String> getVisibleParticipants() {
		// TODO return a copy of the list instead of a reference?
		return visibleParticipants;
	}

	@Override
	public void addRow(IGTDataModel nextRowData) {
		if (nextRowData != null) {
			rowData.add(nextRowData);
			nextRowData.addIGTDataModelListener(this);
		}
	}

	@Override
	public void insertRow(IGTDataModel nextRowData, int row) throws ArrayIndexOutOfBoundsException {
		if (row < 0) {
			throw new ArrayIndexOutOfBoundsException("Insert row, index is less than zero: " +  row);
		} else if (row > rowData.size()){// if insertion point == size, then add 
			throw new ArrayIndexOutOfBoundsException("Insert row, index is greater than model size: " +  row);
		}
		if (rowData != null) {
			rowData.add(row, nextRowData);
			nextRowData.addIGTDataModelListener(this);
		}
	}

	@Override
	public IGTDataModel getRowData(int row) throws ArrayIndexOutOfBoundsException {
		if (row < 0) {
			throw new ArrayIndexOutOfBoundsException("Get row, index is less than zero: " +  row);
		} else if (row > rowData.size() - 1){
			throw new ArrayIndexOutOfBoundsException("Get row, index is greater than model size: " +  row);
		}
		
		return rowData.get(row);
	}	
	
	/**
	 * @see mpi.eudico.client.annotator.interlinear.edit.IGTViewerModel#removeRow(int)
	 */
	@Override
	public IGTDataModel removeRow(int row) throws ArrayIndexOutOfBoundsException {
		if (row < 0) {
			throw new ArrayIndexOutOfBoundsException("Remove row, index is less than zero: " +  row);
		} else if (row > rowData.size() - 1){
			throw new ArrayIndexOutOfBoundsException("Remove row, index is greater than model size: " +  row);
		}
		
		IGTDataModel remRowData = rowData.remove(row);
		if (remRowData != null) {
			remRowData.removeIGTDataModelListener(this);
		}
		
		return remRowData;
	}

	/**
	 * @see mpi.eudico.client.annotator.interlinear.edit.IGTViewerModel#removeRowData(mpi.eudico.client.annotator.interlinear.edit.model.IGTDataModel)
	 */
	@Override
	public boolean removeRowData(IGTDataModel remRowData) {
		if (remRowData == null) {
			return false;
		}
		
		boolean removed = rowData.remove(remRowData);
		if (removed) {
			remRowData.removeIGTDataModelListener(this);
		}
		
		return removed;
	}

	/**
	 * Adds a listener to the model.
	 * 
	 * @param listener
	 */
	@Override
	public synchronized void addIGTViewerModelListener(IGTViewerModelListener listener) {
		if (!listeners.contains(listener)) {
			listeners.add(listener);
		}
	}
	
	/**
	 * Removes the listener from the listeners list.
	 * 
	 * @param listener
	 */
	@Override
	public synchronized void removeIGTViewerModelListener(IGTViewerModelListener listener) {
		listeners.remove(listener);
	}

	// ######## IGTDataModelListener implementation ###################
	/**
	 * If data changed in the one or more of the rows, the viewer model can notify its listeners.
	 * 
	 * @param event the IGT data model event
	 */
	@Override
	public void dataModelChanged(IGTDataModelEvent event) {
		if (event != null) {
			IGTDataModel rowModel = (IGTDataModel) event.getSource();
			int rowIndex = rowData.indexOf(rowModel);
			if (rowIndex > -1) {
				IGTViewerModelEvent vmEvent = new IGTViewerModelEvent(this, rowIndex, event);
				// post event, could have method for that
				for (IGTViewerModelListener l : listeners) {
					l.viewerModelChanged(vmEvent);
				}
			} else {
				// not in the model (yet)?
			}
		}
	}
}
