package mpi.eudico.client.annotator.interlinear.edit;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;

import javax.swing.JComponent;
import javax.swing.SwingConstants;

import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionModel;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionViewerModel;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTTier;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTViewerRenderInfo;

/**
 * A component that visualizes suggestion sets and allows to select one of the suggestions.
 * 
 * @author Han Sloetjes
 *
 */
public class SuggestionSetSelector extends JComponent {
	private IGTSuggestionViewerModel viewerModel;
	private int layoutMode = SwingConstants.HORIZONTAL;
	private IGTGroupRenderer renderer;
	private IGTSuggestionRenderer suggestionRenderer;
	private IGTViewerRenderInfo viewerRenderInfo;
	
	private int selectorWidth;
	private int selectorHeight;
	//private int suggestionMargin = 8;
	
	/**
	 * Constructor accepting suggestions sets in a viewer model as parameter.
	 * 
	 * @param viewerModel the model containing the suggestion sets encapsulated in a igt model
	 * @param viewerRenderInfo contains information for the renderer
	 */
	public SuggestionSetSelector(IGTSuggestionViewerModel viewerModel, IGTViewerRenderInfo viewerRenderInfo) {
		super();
		this.viewerModel = viewerModel;
		this.viewerRenderInfo = viewerRenderInfo;
		renderer = new IGTGroupRenderer();
		suggestionRenderer = new IGTSuggestionRenderer();
		
//		initComponents();
	}
	
	/**
	 * Constructor accepting the list of suggestions as parameter.
	 * 
	 * @param viewerModel the model containing the suggestion sets encapsulated in a igt model
	 * @param viewerRenderInfo contains information for the renderer
	 * @param layoutMode specifies the way the suggestion are presented, horizontally or vertically
	 */
	public SuggestionSetSelector(IGTSuggestionViewerModel viewerModel, IGTViewerRenderInfo viewerRenderInfo, int layoutMode) {
		super();
		this.viewerModel = viewerModel;
		this.viewerRenderInfo = viewerRenderInfo;
		renderer = new IGTGroupRenderer();
		suggestionRenderer = new IGTSuggestionRenderer();
		
		if (layoutMode == SwingConstants.HORIZONTAL || layoutMode == SwingConstants.VERTICAL) {
			this.layoutMode = layoutMode;
		}
//		initComponents();
	}
	
	/**
	 * Call init component after the component is visible, otherwise the graphics object is null.
	 */
	public void init() {
		initComponents();
	}

	/**
	 * Calculates the total height and width of all suggestion sets.
	 */
	private void initComponents() {
		
		if (viewerModel != null) {
			IGTSuggestionModel sugModel;
			
			for (int i = 0; i < viewerModel.getRowCount(); i++) {
				sugModel = viewerModel.getRowData(i);
				
				// hier... calculate sizes
				calculateLocationAndSize(sugModel);
				
				switch(layoutMode) {
				case SwingConstants.VERTICAL:
					selectorHeight += sugModel.getRenderInfo().height;
					selectorHeight += viewerModel.getRenderInfo().suggestionMargin;
					selectorWidth = sugModel.getRenderInfo().width + viewerRenderInfo.headerWidth > selectorWidth ? 
							sugModel.getRenderInfo().width + viewerRenderInfo.headerWidth : selectorWidth;
					break;
					default:// HORIZONTAL
					selectorHeight = sugModel.getRenderInfo().height + viewerModel.getRenderInfo().suggestionMargin > 
						selectorHeight ? sugModel.getRenderInfo().height + viewerModel.getRenderInfo().suggestionMargin : 
							selectorHeight;
					selectorWidth += sugModel.getRenderInfo().width;
					selectorWidth += viewerModel.getRenderInfo().suggestionMargin;
					//if (i == 0) {
						selectorWidth += viewerRenderInfo.headerWidth;
					//}
				}
			}
						
		} else {
			selectorWidth = 150;
			selectorHeight = 100;
		}
		setSize(selectorWidth, selectorHeight);
		setPreferredSize(new Dimension(selectorWidth, selectorHeight));
	}

	/** Paints all suggestion sets in the model.
	 * 
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 * 
	 */
	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2d = (Graphics2D) g;
		
		g2d.setColor(Color.white);
		g2d.fillRect(0, 0, getWidth(), getHeight());
		//IGTSuggestionModel suggestModel;
		
		suggestionRenderer.renderHorizontal(g2d, viewerModel, viewerRenderInfo, 
				0, 0, getWidth(), getHeight());
		/*
		if (viewerModel != null) {
			int curAdv = 0;
			for (int i = 0; i < viewerModel.getRowCount(); i++) {
				suggestModel = viewerModel.getRowData(i);
				int h = suggestModel.getRenderInfo().height;
				int w = suggestModel.getRenderInfo().width;
				
				switch (layoutMode) {
				case SwingConstants.VERTICAL:
		
					renderer.renderRow2(g2d, suggestModel, viewerRenderInfo, i,
							0, curAdv, selectorWidth, h + suggestionMargin);
					
					curAdv += h;
					curAdv += suggestionMargin;
					break;
				case SwingConstants.HORIZONTAL:					
					// hier... alleen de eerste met een "header" sectie?
					renderer.renderRow2(g2d, suggestModel, viewerRenderInfo, i,
							curAdv, 0, w + suggestionMargin, h + viewerRenderInfo.vertRowMargin);
					
					curAdv += w;
					curAdv += viewerRenderInfo.headerWidth;
					if (i != viewerModel.getRowCount() - 1) {
						// visually mark the margin
						g2d.setColor(Color.LIGHT_GRAY);
						g2d.fillRect(curAdv + 1, 0, suggestionMargin - 2, h);
					}
					curAdv += suggestionMargin;
					break;
					
					default:
				}
			}
		}
		*/
	}

	/* (non-Javadoc)
	 * @see javax.swing.JComponent#getPreferredSize()
	 */
	@Override
	public Dimension getPreferredSize() {
		if (selectorHeight == 0 || selectorWidth == 0) {
			return new Dimension(260, 80);
		}
		
		return new Dimension(selectorWidth, selectorHeight);
	}
	
	/**
	 * Calculates the size and location of the items in the model.
	 * 
	 * @param sugModel the suggestion model
	 */
	private void calculateLocationAndSize(IGTSuggestionModel sugModel) {
		Graphics2D g2d = (Graphics2D) getGraphics();

			int h = 0;
			int w = 0;
			if (sugModel.getRowCount() > 0) {
				w = IGTCalculator.calculateTierAndAnnotationsLAS(g2d, viewerRenderInfo, 
						sugModel.getRootRow(), true);
				sugModel.getRenderInfo().width = w;

				IGTTier igtTier = sugModel.getRootRow();
				IGTCalculator.calculateTierYPositionRecursive(g2d, viewerRenderInfo, igtTier, 0);
				
				for (int i = 0; i < sugModel.getRowCount(); i++) {
					igtTier = sugModel.getRowData(i);
					//igtTier.getRenderInfo().height = viewerRenderInfo.tierHeight;
					//h += igtTier.getRenderInfo().height;// TODO get the real tier height
					h += viewerRenderInfo.getHeightForTier(g2d, igtTier.getTierName());
					h += viewerRenderInfo.vertLineMargin;
				}
				h += (2 * viewerRenderInfo.vertLineMargin);
				sugModel.getRenderInfo().height = h;	
			}
	}
	
	/**
	 * Returns the index of the model displayed at the given location.
	 * 
	 * @param x the x coordinate
	 * @param y the y coordinate
	 * @return the index of the model, or -1 if none
	 */
	public int getSuggestionIndexAtPoint(int x, int y) {
		if (viewerModel != null) {
			IGTSuggestionModel suggestModel;
			int advance = 0;
			int w, h;
			
			for (int i = 0; i < viewerModel.getRowCount(); i++) {
				suggestModel = viewerModel.getRowData(i);
				w = suggestModel.getRenderInfo().width;
				h = suggestModel.getRenderInfo().height;
				
				switch(layoutMode) {
				case SwingConstants.VERTICAL:
					if (y >= advance && y < advance + h) {
						return i;
					}
					advance += h;
					break;
					
				case SwingConstants.HORIZONTAL:
					if (i == 0) {
						advance += viewerRenderInfo.headerWidth;
					}
					
					if (x >= advance && x < advance + w) {
						return i;
					}
					advance += w;
					break;
					
					default:
				}
				
				advance += viewerModel.getRenderInfo().suggestionMargin;
			}
		}
		
		return -1;
	}
	
	/**
	 * Returns the suggestion viewer model.
	 * 
	 * @return the suggestion viewer model
	 */
	public IGTSuggestionViewerModel getModel() {
		return viewerModel;
	}
}
