package mpi.eudico.client.annotator.interlinear.edit.model;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import mpi.eudico.client.annotator.interlinear.IGTTierType;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTTierRenderInfo;

/**
 * A tier object representing a row in a IGTDataModel.
 * The tier can contain one or more IGTAnnotations. For "root" tiers or tiers
 * symbolically associated with the root the number of annotations is 0 or 1. 
 * 
 * @author Han Sloetjes
 */
public class IGTTier {
    private String tierName;
    private ArrayList<IGTAnnotation> annotations;
    private IGTTier parentTier;
    private List<IGTTier> childTiers;
	/**
	 * The type of the IGT tier
	 */
	private IGTTierType type;
	private IGTTierRenderInfo renderInfo;
	/** A flag to set if this tier is a "word" level subdivision tier or a descendant tier thereof
	 *  This flag should be consistent with the tier type 
	 *  (could also define specific tier types for such tiers but this flag is a useful shorthand test) */
	private boolean isInWordLevelBlock = false;
    
	/**
	 * Constructor.
	 * 
	 * @param tierName the name of the tier
	 * @param type the type of tier
	 */
	public IGTTier(String tierName, IGTTierType type) {
		this.tierName = tierName;
		this.type = type;
		renderInfo = new IGTTierRenderInfo();
		annotations = new ArrayList<IGTAnnotation>();
		parentTier = null;
		childTiers = new ArrayList<IGTTier>(8);
	}
	
	// some standard getters and setters
	public String getTierName() {
		return tierName;
	}

	public void setTierName(String tierName) {
		this.tierName = tierName;
	}

	public ArrayList<IGTAnnotation> getAnnotations() {
		return annotations;
	}

	public void setAnnotations(ArrayList<IGTAnnotation> annotations) {
		if (annotations != null) {// prevent annotations from being set to null
			this.annotations = annotations;
		}
	}

	public IGTTierType getType() {
		return type;
	}

	public void setType(IGTTierType type) {
		this.type = type;
	}
	
    /**
     * Adds an annotation to this IGTTier.
     *
     * @param ann a <code>IGTAnnotation</code> that is to be added to
     *        this IGTTier
     */
    public void addAnnotation(IGTAnnotation ann) {
        if (ann != null) {
            annotations.add(ann);
            ann.setIGTTier(this);
        }
    }   
    
    /**
     * 
     * 
     * @param ann
     * @param index
     */
    public void insertAnnotation(IGTAnnotation ann, int index) 
    		throws ArrayIndexOutOfBoundsException {
        if (ann != null) {
            annotations.add(index, ann);
            ann.setIGTTier(this);
        }
    }
    
    /**
     * Removes an annotation from the IGTTier.
     * 
     * @param ann 
     */
    public void removeAnnotation(IGTAnnotation ann) {
    	if (ann != null) {
            annotations.remove(ann);
            ann.setIGTTier(null);
        }
    }
    
    /**
     * 
     * @param ann the reference annotation
     * @return the next annotation on this tier, or null
     */
    public IGTAnnotation getNextAnnotation(IGTAnnotation ann) {
    	if (ann == null) {
    		return null;
    	}
    	
    	int index = annotations.indexOf(ann);
    	if (index > -1 && index < annotations.size() - 1) {
    		return annotations.get(index + 1);
    	}
    	
    	// else return null
    	return null;
    }
    
    
    /**
     * 
     * @param ann the reference annotation
     * @return the previous annotation on this tier, or null
     */
    public IGTAnnotation getPreviousAnnotation(IGTAnnotation ann) {
    	if (ann == null) {
    		return null;
    	}
    	
    	int index = annotations.indexOf(ann);
    	if (index > 0) {
    		return annotations.get(index - 1);
    	}
    	
    	// else return null
    	return null;
    }
    
    /**
     * Returns the render information object.
     * 
     * @return the render information object
     */
    public IGTTierRenderInfo getRenderInfo() {
    	return renderInfo;
    }

	/**
	 * @return the parentTier
	 */
	public IGTTier getParentTier() {
		return parentTier;
	}
	
	/**	
	 * @return the root tier 
	 */
	public final IGTTier getRootTier() {
		if (parentTier == null) 
			return this;
		return parentTier.getRootTier();
	}

	/**
	 * @param parentTier the parentTier to set
	 */
	public void setParentTier(IGTTier parentTier) {
		this.parentTier = parentTier;
	}

	/**
	 * This method returns the direct children of this tier. 
	 * This list does not contain deeper level descendants (grand
	 * children).  
	 * 
	 * @return the childTiers
	 */
	public List<IGTTier> getChildTiers() {
		return childTiers;
	}
	
	/**
	 * Returns a list of descendant tiers. The tiers are added recursively, depth first.
	 * 
	 * @return a list of descendant tiers, hierarchically
	 */
	public List<IGTTier> getDescendantTiers() {
		ArrayList<IGTTier> descendants = new ArrayList<IGTTier>();
		addDescendantsToList(descendants);
		
		return descendants;
	}
	
	/**
	 * Adds children to the list, recursively.
	 * 
	 * @param descendants the list to add to
	 */
	private void addDescendantsToList(ArrayList<IGTTier> descendants) {
		if (descendants == null) {
			return;
		}
		IGTTier t;
		for (int i = 0; i < childTiers.size(); i++) {
			t = childTiers.get(i);
			descendants.add(t);
			t.addDescendantsToList(descendants);
		}
	}
    
	/**
	 * Adds a child tier to the list of child tiers.
	 *  
	 * @param childTier the child tier to add
	 */
	public void addChildTier(IGTTier childTier) {
		if (childTier != null && !childTiers.contains(childTier)) {
			childTiers.add(childTier);
		}
	}
	
	/**
	 * Removes the given child tier
	 *  
	 * @param childTier the child tier to remove
	 */
	public void removeChildTier(IGTTier childTier){
		if (childTier != null && childTiers.contains(childTier)) {
			childTiers.remove(childTier);
		}
	}
	
	/**
	 * Returns whether this tier is a descendant of the specified tier, i.e. whether 
	 * the other tier is an ancestor of this tier.
	 * 
	 * @param otherTier the possible ancestor tier
	 * @return true if the other tier is an ancestor of this tier, false otherwise
	 */
	public boolean hasAncestor(IGTTier otherTier) {
		if (otherTier == null || otherTier == this) {
			return false;
		}
		
		IGTTier curParent = parentTier;
		while (curParent != null) {
			if (curParent == otherTier) {
				return true;
			}
			
			curParent = curParent.getParentTier();
		}
		
		return false;
	}
	
	/**
	 * Returns whether this tier is part of an interlinear block, i.e.
	 * a tier that is a "word-level" subdivision tier or a descendant thereof.
	 *  
	 * @return true if this tier is in an interlinear block
	 */
	public boolean isInWordLevelBlock() {
		return isInWordLevelBlock;
	}
	
	/**
	 * Sets the flag whether this tier is part of an interlinear block.
	 * This should be consistent with the type of this tier.
	 * 
	 * @param inBlock the new value of the flag
	 */
	public void setInWordLevelBlock(boolean inBlock) {
		if (inBlock) {
			if (type != IGTTierType.ROOT && type != IGTTierType.PHRASE_LEVEL_ROOT && type != IGTTierType.FIRST_LEVEL_ASSOCIATION && 
					type != IGTTierType.SILENCE_DURATION && type != IGTTierType.SPEAKER_LABEL && 
					type != IGTTierType.TIME_CODE) {
				isInWordLevelBlock = inBlock;
			} else {
				// don't set the flag, throw an exception?
			}
		} else {
			if (type == IGTTierType.ROOT || type == IGTTierType.PHRASE_LEVEL_ROOT || type == IGTTierType.FIRST_LEVEL_ASSOCIATION || 
					type == IGTTierType.SILENCE_DURATION || type == IGTTierType.SPEAKER_LABEL || 
					type == IGTTierType.TIME_CODE) {// allow
				isInWordLevelBlock = inBlock;
			} else {
				// don't set the flag, throw an exception?
			}
		}
	}
	
	// #### methods for locating annotations etc #############
	/**
	 * Returns whether this tier has any, possibly wrapped, line at the specified y location. 
	 * 
	 * @param y the y location
	 * @return true if this tier has a line at the y coordinate, false otherwise
	 */
	public boolean isAtY(int y) {
		if (renderInfo.numLines == 1) {
			return y >= renderInfo.y && y <= renderInfo.y + renderInfo.height;
		} else {
			// TODO check the grid of wrapped lines / blocks
			if (isInWordLevelBlock) {
				List<Integer> yPos = renderInfo.getYPositions(); 
				if (yPos != null) {
					int nextY;
					for (int i = 0; i < yPos.size(); i++) {
						nextY = yPos.get(i);
						if (y >= nextY && y <= nextY + renderInfo.height) {
							return true;
						}
						if (nextY > y) {
							return false;
						}
					}
				}
			} else /*if (type == IGTTierType.ROOT || type == IGTTierType.FIRST_LEVEL_ASSOCIATION || 
					type == IGTTierType.DEEPER_LEVEL_ASSOCIATION)*/ {
				return y >= renderInfo.y && y <= renderInfo.y + renderInfo.renderHeight;
			}
		}
		return false;
	}
	
	/**
	 * Returns the annotation that is located at the specified point given the current rendering settings.
	 * 
	 * @param p the point to query 
	 * @return the annotation at the point or null
	 */
	public IGTAnnotation getAnnotationAtPoint(Point p) {
		
		if (!isInWordLevelBlock) {
			// exclude special tiers, like time code, speaker, silence tier?
			if (annotations.size() > 0) {
				IGTAnnotation igtAnn = annotations.get(0);
				if (igtAnn.getRenderInfo().isPointInRenderArea(p)) {
					return igtAnn;
				}
			}
		} else {
			IGTAnnotation igtAnn;
			if (getRenderInfo().getWrapIndices() != null) {
				int wrapIndex = 0;
				
				for (int i = 0; i < annotations.size(); i++) {
					igtAnn = annotations.get(i);
					if (igtAnn.getRenderInfo().isPointInRenderArea(p)) {
						return igtAnn;
					}
					if (igtAnn.getRenderInfo().y + igtAnn.getRenderInfo().height < p.y) {
						if (wrapIndex < getRenderInfo().getWrapIndices().size()) {
							i = getRenderInfo().getWrapIndices().get(wrapIndex);
							wrapIndex++;
						}
					}
				}
			} else {
				// single line of annotations
				for (int i = 0; i < annotations.size(); i++) {
					igtAnn = annotations.get(i);
					if (igtAnn.getRenderInfo().isPointInRenderArea(p)) {
						return igtAnn;
					}
				}
			}
		}
		
		return null;
	}
}
