package mpi.eudico.client.annotator.layout;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;

import javax.swing.JButton;
import javax.swing.SwingConstants;
import javax.swing.border.SoftBevelBorder;

import mpi.eudico.client.annotator.ElanLayoutManager;
import mpi.eudico.client.annotator.ElanMediaPlayerController;
import mpi.eudico.client.annotator.ModePanel;
import mpi.eudico.client.annotator.Preferences;
import mpi.eudico.client.annotator.SelectionButtonsPanel;
import mpi.eudico.client.annotator.ViewerManager2;
import mpi.eudico.client.annotator.commands.ELANCommandFactory;
import mpi.eudico.client.annotator.gui.ResizeComponent;
import mpi.eudico.client.annotator.turnsandscenemode.TaSFontSizePanel;
import mpi.eudico.client.annotator.turnsandscenemode.TurnsAndSceneViewer;
import mpi.eudico.client.annotator.util.FrameConstants;
import mpi.eudico.client.annotator.viewer.SignalViewer;
import mpi.eudico.server.corpora.clom.Transcription;

/**
 * Layout manager for a simplified layout of a transcription window,
 * with at most one video, with (at most) one waveform viewer and one 
 * area that is an annotation viewer/editor for (initially) one tier. 
 * A segmentation is added by hitting the Enter key.
 * 
 * Eventually it would be nice if the video area can be either WEST, EAST, NORTH or SOUTH
 * and the waveform viewer either NORTH or SOUTH.
 */
public class TurnsAndSceneManager implements ModeLayoutManager {
	private ElanLayoutManager layoutManager;
    private Container container;
    private ViewerManager2 viewerManager; 
    private Transcription transcription;
    private SignalViewer signalViewer;
    private TurnsAndSceneViewer turnsAndSceneViewer;
	private ResizeComponent horMediaResizer;
	private ResizeComponent vertViewerResizer;
	private ElanMediaPlayerController mediaPlayerController;
	
	private TaSFontSizePanel fontSizePanel;
	
	private boolean videoInRight = false;
	private boolean initialized = false;
	private int numVisualPlayers;
	
	private final int MIN_SIGNAL_HEIGHT = 100;
	
	private final int MIN_VIEWER_HEIGHT = 120;

	private JButton butAlignVideoToRight;

	
	/**
	 * @param viewerManager the viewer manager
	 * @param elanLayoutManager the layout manager
	 */
	public TurnsAndSceneManager(ViewerManager2 viewerManager, ElanLayoutManager elanLayoutManager) {
		this.viewerManager = viewerManager;
        this.layoutManager = elanLayoutManager;
        transcription = viewerManager.getTranscription();        
        if(viewerManager.getMasterMediaPlayer() != null){
        	viewerManager.getMasterMediaPlayer().setMediaTime(0);
        }
        
        container = layoutManager.getContainer();         
	}

	@Override
	public void add(Object object) {
		if (object instanceof SignalViewer) {
	    	setSignalViewer((SignalViewer) object);
		} else if (object instanceof TurnsAndSceneViewer) {
			setSingleLayerViewer((TurnsAndSceneViewer) object);
		} else if (object instanceof ElanMediaPlayerController) {
			this.mediaPlayerController = (ElanMediaPlayerController) object;
			container.add(mediaPlayerController.getPlayButtonsPanel());
			SelectionButtonsPanel sbp = mediaPlayerController.getSelectionButtonsPanel();
			sbp.setSeparateLeftRightMode(true);
			container.add(sbp);
			container.add(mediaPlayerController.getTimePanel());
			container.add(mediaPlayerController.getModePanel());
			((ModePanel) mediaPlayerController.getModePanel()).setModeVisible(ELANCommandFactory.SELECTION_MODE, false);
//			for (Component button : mediaPlayerController.getPlayButtonsPanel().getComponents()) {
//				button.setFocusable(false);
//			}
		} else if (object instanceof TaSFontSizePanel) {
			fontSizePanel = (TaSFontSizePanel) object;
			container.add(fontSizePanel);
		}
	}

	/**
	 * This still has to be implemented
	 */
	@Override
	public void remove(Object object) {
		// TODO implement
		
	}
	
	private void setSignalViewer(SignalViewer sigViewer){
		this.signalViewer = sigViewer;	
		signalViewer.setMediaTime(0);
		signalViewer.setSelection(0, 0);
		container.add(signalViewer);
        if (initialized) { 
        	doLayout();
        }  
	}
	
	private void setSingleLayerViewer(TurnsAndSceneViewer slViewer) {
		this.turnsAndSceneViewer = slViewer;
		container.add(turnsAndSceneViewer);
		
		if (initialized) { 
        	doLayout();
        }
	}
	

	/**
	 * TODO 
	 * - consider to allow to place the video etc. on the right side
	 * - if no video, place the controls somewhere else, e.g. between the annotation viewer and the signal viewer
	 * - add loop mode
	 * - add a scroll bar for the signal viewer?
	 * - add volume and rate slider?, or configuration items
	 */
	@Override
	public void doLayout() {
		if (!initialized) {
			return;
		}
		if (layoutManager.getAttachedVisualPlayers().length == 0) {
			doLayoutNoVideo();
		} else {
			doLayoutWithVideo();
		}
	}
	
	
	private void doLayoutWithVideo() {
		if (!initialized) {
			return;
		}
		
		// place one video in the left corner
		int containerWidth 		= container.getWidth();
	    int containerHeight 	= container.getHeight();
	    int containerMargin 	= 3;
	    int componentMargin 	= 5;
	    
	    int mediaAreaWidth 		= layoutManager.getMediaAreaWidth();
	    int mediaAreaHeight 	= layoutManager.getMediaAreaHeight();// re-use for now?
	    int visibleMediaX 		= containerMargin;
	    int visibleMediaY 		= containerMargin;	
	    // ensure a minimal height for the signal viewer
	    if (signalViewer == null) {
	    	mediaAreaHeight = containerHeight;
	    } else if (mediaAreaHeight > containerHeight - MIN_SIGNAL_HEIGHT) {
	    	mediaAreaHeight = containerHeight - MIN_SIGNAL_HEIGHT;
	    }
	    int curSignalHeight = containerHeight - mediaAreaHeight;
	    // ensure a minimal width for the player buttons
	    if(mediaPlayerController != null) {
	    	int minWidth = mediaPlayerController.getPlayButtonsPanel().getPreferredSize().width;
	    	mediaAreaWidth = mediaAreaWidth < minWidth ? minWidth : mediaAreaWidth;
	    }
	    
	    PlayerLayoutModel[] visualPlayers = layoutManager.getAttachedVisualPlayers();
		numVisualPlayers = visualPlayers.length;
		
		// if there is no media
		//if(numVisualPlayers == 0){// TODO remove if layout without video has been implemented
			if(videoInRight){
				visibleMediaX = containerWidth - (mediaAreaWidth + containerMargin);
			} 
		//} 
		
		if (numVisualPlayers >= 1) {
			// layout the first video
			Component firstVisualComp = visualPlayers[0].visualComponent;
			float aspectRatio = visualPlayers[0].player.getAspectRatio();
			int firstMediaHeight = (int) (mediaAreaWidth / aspectRatio);				
			if(videoInRight){	
				visibleMediaX = (containerWidth - (mediaAreaWidth + containerMargin));
			}			
			firstVisualComp.setBounds(visibleMediaX, visibleMediaY, mediaAreaWidth,
						firstMediaHeight);	
			visibleMediaY = visibleMediaY + firstMediaHeight;
		}

		// layout for the time panel
		int timePanelX = visibleMediaX;
		int timePanelY = visibleMediaY + 2;
		int timePanelWidth = 0;
		int timePanelHeight = 0;        
		if (mediaPlayerController != null) {
			timePanelWidth = mediaPlayerController.getTimePanel().getPreferredSize().width;
			timePanelHeight = mediaPlayerController.getTimePanel().getPreferredSize().height;
			if(numVisualPlayers > 0){	
				timePanelX = (visibleMediaX + (mediaAreaWidth / 2)) - (timePanelWidth / 2);
			} 					
			mediaPlayerController.getTimePanel().setBounds(timePanelX, timePanelY, timePanelWidth, timePanelHeight);
		}
        
		//layout for the video Alignment button
//		int butAlignVideoX = visibleMediaX;
//		int butAlignVideoY = timePanelY + timePanelHeight + 4;
//		int butAlignVideoWidth = butAlignVideoToRight.getPreferredSize().width;
//		int butAlignVideoHeight = butAlignVideoToRight.getPreferredSize().height;       
//		butAlignVideoToRight.setBounds(butAlignVideoX, butAlignVideoY, butAlignVideoWidth, butAlignVideoHeight);
		// place control buttons right below the video,
		// layout for the buttons in 2 or 3 rows
		int playButtonX = visibleMediaX;
		int playButtonY = timePanelY + timePanelHeight + 4;
		int playButtonWidth = 0;
		int playButtonHeight = 0;
		int playSelectionButtonX = playButtonX;
		int playSelectionButtonY = playButtonY;
		if(mediaPlayerController != null){
			playButtonWidth = mediaPlayerController.getPlayButtonsPanel().getPreferredSize().width;
			playButtonHeight = mediaPlayerController.getPlayButtonsPanel().getPreferredSize().height;
	        //if(numVisualPlayers > 0 ) {	
			playButtonX = (mediaAreaWidth - playButtonWidth) / 2;	
			//}

			mediaPlayerController.getPlayButtonsPanel().setBounds(playButtonX, playButtonY, playButtonWidth, playButtonHeight);
			if (numVisualPlayers == 0) {
				// center-align the time panel with the button panel
				int mid = playButtonX + (playButtonWidth / 2);
				mediaPlayerController.getTimePanel().setBounds(mid - timePanelWidth / 2, timePanelY, timePanelWidth, timePanelHeight);
			}
			// layout selection button panel beneath the player buttons, align left
			playSelectionButtonX = playButtonX /*+ ((playButtonWidth - mediaPlayerController.getSelectionButtonsPanel().getPreferredSize().width) / 2)*/;
			playSelectionButtonY += (playButtonHeight + componentMargin);
			mediaPlayerController.getSelectionButtonsPanel().setBounds(playSelectionButtonX, playSelectionButtonY, 
					mediaPlayerController.getSelectionButtonsPanel().getPreferredSize().width, 
					mediaPlayerController.getSelectionButtonsPanel().getPreferredSize().height);
			// layout mode panel at the same y-location as selection buttons, but right aligned to player buttons panel?
			int modePanelWidth = mediaPlayerController.getModePanel().getPreferredSize().width;
			int modePanelHeight = mediaPlayerController.getModePanel().getPreferredSize().height;
			mediaPlayerController.getModePanel().setBounds(playButtonX + playButtonWidth - modePanelWidth, 
					playSelectionButtonY, modePanelWidth, modePanelHeight);
			
			// overrule the minimal signal viewer height
			if (mediaAreaHeight < playSelectionButtonY + mediaPlayerController.getSelectionButtonsPanel().getPreferredSize().height) {
				mediaAreaHeight = playSelectionButtonY + mediaPlayerController.getSelectionButtonsPanel().getPreferredSize().height;
			}
		}
		
		if (fontSizePanel != null) {
			int fontPanelX = visibleMediaX;
			int fontPanelY = playSelectionButtonY + playButtonHeight +  2 * containerMargin;
			int fontPanelWidth = fontSizePanel.getPreferredSize().width;
			int fontPanelHeight = fontSizePanel.getPreferredSize().height;
			
			if (playButtonWidth != 0) {
				fontPanelX = playButtonX + (playButtonWidth - fontPanelWidth) / 2;
			}
			// place at bottom
			if (mediaAreaHeight > fontPanelY + fontPanelHeight) {
				fontPanelY = mediaAreaHeight - fontPanelHeight;
			}
			fontSizePanel.setBounds(fontPanelX, fontPanelY, fontPanelWidth, fontPanelHeight);
			
			if (mediaAreaHeight < fontPanelY + fontPanelHeight) {
				mediaAreaHeight = fontPanelY + fontPanelHeight;
			}
		}
		// place the vertical viewer resizer
        int divX = visibleMediaX + mediaAreaWidth + containerMargin; // always keep visible
        int divY = 0; 
        int divWidth = horMediaResizer.getPreferredSize().width;     
        if (videoInRight){
        	divX = visibleMediaX - containerMargin - divWidth;
        } 
        //use mediaAreaheight
        horMediaResizer.setBounds(divX, divY, divWidth, mediaAreaHeight);
		
        if (signalViewer != null) { // check if there is a SignalViewer
	        // place the horizontal resizer, based on the height of the signal viewer / or on the media area height
	        
	        int horDivX = 0;
			int horDivY = mediaAreaHeight;
			int horDivHeight = vertViewerResizer.getPreferredSize().height;
			vertViewerResizer.setBounds(horDivX, horDivY, containerWidth, horDivHeight);
			// place the signal viewer at the bottom (or top) of the window
			int signalX = visibleMediaX;// or 0
			int signalY = horDivY + horDivHeight;
			int signalW = containerWidth;// margins?
			int signalH = curSignalHeight;
			signalViewer.setBounds(signalX, signalY, signalW, signalH);
        } else {
        	// remove the resizer, fix it at the bottom?
        	
        }
		// place the simple transcription viewer in the remaining area
        int stViewX = divX + divWidth + containerMargin;
        int stViewY = containerMargin;
        int stViewW = containerWidth - containerMargin - stViewX;
        int stViewH = mediaAreaHeight;
        
        turnsAndSceneViewer.setBounds(stViewX, stViewY, stViewW, stViewH);
	}

	/**
	 * No video means there is no divider between the video + controls and the annotation viewer.
	 * The player control buttons and other buttons are positioned below the viewer and above the
	 * signal viewer (if present).
	 */
	private void doLayoutNoVideo() {
		if (!initialized) {
			return;
		}
		/*
	    PlayerLayoutModel[] visualPlayers = layoutManager.getAttachedVisualPlayers();
		numVisualPlayers = visualPlayers.length;
		
		if (numVisualPlayers > 0) {
			return;
		}
		*/
		// place one video in the left corner
		int containerWidth 		= container.getWidth();
	    int containerHeight 	= container.getHeight();
	    int containerMargin 	= 3;
	    int componentMargin 	= 5;
	    
	    int mediaAreaHeight 	= layoutManager.getMediaAreaHeight();// re-use for now - is total height minus signal viewer height
	    int tasViewerX  		= containerMargin;
	    int tasViewerY  		= containerMargin;
	    int tasViewerWidth = containerWidth - (2 * containerMargin);
	    int controlsAreaWidth   = 0;
	    int controlsAreaHeight  = 0;
	    int controlsAreaY = MIN_VIEWER_HEIGHT;
	    // ensure a minimal height for the signal viewer
	    if (signalViewer == null) {
	    	mediaAreaHeight = containerHeight;
	    } else if (mediaAreaHeight > containerHeight - MIN_SIGNAL_HEIGHT) {
	    	mediaAreaHeight = containerHeight - MIN_SIGNAL_HEIGHT;
	    }
	    
	    // ensure a minimal width for the player buttons
		// layout for the time panel
		int timePanelX = tasViewerX;
		int timePanelY;
		int timePanelWidth = 0;
		int timePanelHeight = 0; 
		int playButtonX = tasViewerX;
		int playButtonY;
		int playButtonWidth = 0;
		int playButtonHeight = 0;
		int playSelectionButtonX = playButtonX;
		int playSelectionButtonY;
		int playSelectionButtonWidth= 0;
		int playSelectionButtonHeight = 0;
		
	    if(mediaPlayerController != null) {
			// place control buttons to the left in the window, below the annotation viewer
			timePanelWidth = mediaPlayerController.getTimePanel().getPreferredSize().width;
			timePanelHeight = mediaPlayerController.getTimePanel().getPreferredSize().height;
			playButtonWidth = mediaPlayerController.getPlayButtonsPanel().getPreferredSize().width;
			playButtonHeight = mediaPlayerController.getPlayButtonsPanel().getPreferredSize().height;
	    	controlsAreaWidth = controlsAreaWidth < playButtonWidth ? playButtonWidth : controlsAreaWidth;
	    	playSelectionButtonWidth = mediaPlayerController.getSelectionButtonsPanel().getPreferredSize().width;
	    	playSelectionButtonHeight = mediaPlayerController.getSelectionButtonsPanel().getPreferredSize().height;

			int modePanelWidth = mediaPlayerController.getModePanel().getPreferredSize().width;
			int modePanelHeight = mediaPlayerController.getModePanel().getPreferredSize().height;
			int thirdRowHeight = playSelectionButtonHeight > modePanelHeight ? playSelectionButtonHeight : modePanelHeight;// the row for play selection and loop mode
			
			controlsAreaHeight = (4 * componentMargin) + timePanelHeight + playButtonHeight + thirdRowHeight;
			controlsAreaY = mediaAreaHeight - controlsAreaHeight;
			if (controlsAreaY < MIN_VIEWER_HEIGHT) {
				controlsAreaY = MIN_VIEWER_HEIGHT;// let min viewer height override min signal viewer height
				mediaAreaHeight = controlsAreaY + controlsAreaHeight;
			}
			// center-align the time panel with the button panel
			int playButtonMid = playButtonX + (playButtonWidth / 2);
			timePanelX = playButtonMid - (timePanelWidth / 2);
			timePanelY = controlsAreaY + componentMargin;
			mediaPlayerController.getTimePanel().setBounds(timePanelX, timePanelY, timePanelWidth, timePanelHeight);
			playButtonY = timePanelY + timePanelHeight + componentMargin;
			mediaPlayerController.getPlayButtonsPanel().setBounds(playButtonX, playButtonY, 
					playButtonWidth, playButtonHeight);
			
			// layout selection button panel beneath the player buttons, align left
			playSelectionButtonX = playButtonX;
			playSelectionButtonY = playButtonY + playButtonHeight + componentMargin;
			if (thirdRowHeight > playSelectionButtonY) {
				playSelectionButtonY += ((thirdRowHeight - playSelectionButtonY) / 2);
			}
			mediaPlayerController.getSelectionButtonsPanel().setBounds(playSelectionButtonX, playSelectionButtonY, 
					playSelectionButtonWidth, playSelectionButtonHeight);
			// layout mode panel at the same y-location as selection buttons, but right aligned to player buttons panel?
			int modePanelY = playButtonY + playButtonHeight + componentMargin;
			if (modePanelHeight < thirdRowHeight) {
				modePanelY += ((thirdRowHeight - playSelectionButtonY) / 2);
			}
			mediaPlayerController.getModePanel().setBounds(playButtonX + playButtonWidth - modePanelWidth, 
					modePanelY, modePanelWidth, modePanelHeight);
		}
		
	    
		if (fontSizePanel != null) {
			int fontPanelX;
			int fontPanelY = controlsAreaY + componentMargin;
			int fontPanelWidth = fontSizePanel.getPreferredSize().width;
			int fontPanelHeight = fontSizePanel.getPreferredSize().height;
			fontPanelX = containerWidth - containerMargin - fontPanelWidth;
			// maybe check that fontPanelX > playButtonX + playButtonWidth + componentMargin;
			if (mediaPlayerController == null) {
				controlsAreaHeight = fontPanelHeight + (2 * componentMargin);
				fontPanelY = mediaAreaHeight - fontPanelHeight - (2 * componentMargin);
			}
			
			fontSizePanel.setBounds(fontPanelX, fontPanelY, fontPanelWidth, fontPanelHeight);
			//fontSizePanel.setVisible(fontPanelX > playButtonX + playButtonWidth);
		}
		
		int tasViewerHeight = controlsAreaY - containerMargin;
	    
        if (signalViewer != null) { // check if there is a SignalViewer
	        // place the horizontal resizer, based on the height of the signal viewer / or on the media area height   
	        int horDivX = 0;
			int horDivY = mediaAreaHeight;
			int horDivHeight = vertViewerResizer.getPreferredSize().height;
			vertViewerResizer.setBounds(horDivX, horDivY, containerWidth, horDivHeight);
			// place the signal viewer at the bottom (or top) of the window
			int signalX = 0;// or tasViewerX = containerMargin
			int signalY = horDivY + horDivHeight;
			int signalW = containerWidth;// margins?
			int signalH = containerHeight - mediaAreaHeight - horDivHeight;
			signalViewer.setBounds(signalX, signalY, signalW, signalH);
        } else {
        	// remove the resizer, fix it at the bottom?
        	
        }
		// place the simple transcription viewer in the remaining area, at the top
        turnsAndSceneViewer.setBounds(tasViewerX, tasViewerY, tasViewerWidth, tasViewerHeight);
	}
	
	@Override
	public void updateLocale() {
		// TODO implement? call updateLocale on all items that are not directly connected as locale listeners
	}

	@Override
	public void clearLayout() {
		// TODO implement
		
	}

	@Override
	public void initComponents() {
		 // divider component TODO change this: create and add when the first video is added
        horMediaResizer = new ResizeComponent(layoutManager, SwingConstants.HORIZONTAL);
        horMediaResizer.setBorder(new SoftBevelBorder(SoftBevelBorder.RAISED)); 
        horMediaResizer.setPreferredSize(new Dimension(7,container.getHeight()));
        Component n = horMediaResizer.getComponent(0);
        horMediaResizer.remove(n);
        horMediaResizer.setLayout(new GridBagLayout());
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.anchor = GridBagConstraints.PAGE_END;
        gbc.weighty = 1.0; 
        horMediaResizer.add(n, gbc);
        
    	container.add(horMediaResizer);
    	// TODO create and add this when the signal viewer is added
    	vertViewerResizer = new ResizeComponent(layoutManager, SwingConstants.VERTICAL);
    	vertViewerResizer.setBorder(new SoftBevelBorder(SoftBevelBorder.RAISED)); 
    	vertViewerResizer.setPreferredSize(new Dimension(container.getWidth(), 7));
   	
    	n = vertViewerResizer.getComponent(0);
    	vertViewerResizer.remove(n);
    	vertViewerResizer.setLayout(new GridBagLayout());
        gbc = new GridBagConstraints();
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.weightx = 1.0; 
        vertViewerResizer.add(n, gbc);
    	
    	container.add(vertViewerResizer);
    	butAlignVideoToRight = new JButton("VL");
    	//container.add(butAlignVideoToRight);
    	
    	add(viewerManager.getMasterMediaPlayer());
    	add(viewerManager.getMediaPlayerController());
    	layoutManager.add(viewerManager.createSignalViewer());
		add(viewerManager.createTurnsAndSceneViewer());
		if (turnsAndSceneViewer != null) {
			add(new TaSFontSizePanel(turnsAndSceneViewer));
		}
		initialized = true;
		doLayout();
	}

	/**
	 * Maybe this can/should be implemented. It should be implemented in case this mode should ever
	 * become one of the modes in ELAN (rather than the single mode of a different application).
	 */
	@Override
	public void enableOrDisableMenus(boolean enabled) {
//		((ElanFrame2)layoutManager.getElanFrame()).enableOrDisableMenus(null, FrameConstants.ANNOTATION, enabled);
//		((ElanFrame2)layoutManager.getElanFrame()).enableOrDisableMenus(null, FrameConstants.TIER, enabled);
//		((ElanFrame2)layoutManager.getElanFrame()).enableOrDisableMenus(null, FrameConstants.TYPE, enabled);
//		((ElanFrame2)layoutManager.getElanFrame()).enableOrDisableMenus(null, FrameConstants.SEARCH, enabled);
//		((ElanFrame2)layoutManager.getElanFrame()).enableOrDisableMenus(null, FrameConstants.VIEW, enabled);
//		((ElanFrame2)layoutManager.getElanFrame()).enableOrDisableMenus(null, FrameConstants.OPTION, enabled);

	}

	/**
	 * This should be implemented if this layout became one of the modes in ELAN. Now it is not possible to change
	 * shortcut keys.
	 */
	@Override
	public void shortcutsChanged() {
//		loadKSNottoBeConsumed();
		// update viewer and buttons
	}

	@Override
	public void createAndAddViewer(String viewerName) {
		// TODO maybe use this?
		
	}

	/**
	 * @return true if the viewer was there and has been removed (and repaint is needed)
	 */
	@Override
	public boolean destroyAndRemoveViewer(String viewerName) {
		// TODO implement if necessary
		return false;
	}

	@Override
	public void detach(Object object) {
		// TODO implement if necessary
		
	}

	@Override
	public void attach(Object object) {
		// TODO implement if necessary
		
	}

	@Override
	public void preferencesChanged() {
		Integer maHeight = Preferences.getInt("MediaAreaHeight-SL", transcription);
		if (maHeight == null) {
			int height = container.getHeight();
			if (height > 0 && height - layoutManager.getMediaAreaHeight() > MIN_SIGNAL_HEIGHT) {
				layoutManager.setMediaAreaHeight(height - MIN_SIGNAL_HEIGHT);
			}
		} else {
			layoutManager.setMediaAreaHeight(maHeight);
		}
		Integer maWidth = Preferences.getInt("MediaAreaWidth-SL", transcription);
		if (maWidth != null) {
			layoutManager.setMediaAreaWidth(maWidth);
		}
		Integer fontSize = Preferences.getInt("TurnsAndSceneMode.View.FontSize", transcription);
		if (fontSize != null && turnsAndSceneViewer != null) {
			turnsAndSceneViewer.setFontSize(fontSize);
		}
				
		initialized = true;
	}

	@Override
	public void cleanUpOnClose() {
		// TODO implement proper cleanup and disconnection of objects
	}

	@Override
	public void isClosing() {
		layoutManager.setPreference("MediaAreaHeight-SL", layoutManager.getMediaAreaHeight(), transcription);
		layoutManager.setPreference("MediaAreaWidth-SL", layoutManager.getMediaAreaWidth(), transcription);
		layoutManager.setPreference("TurnsAndSceneMode.View.FontSize", turnsAndSceneViewer.getFontSize(), transcription);
	}



}
