package mpi.eudico.client.annotator.turnsandscenemode;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Rectangle;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JRootPane;
import javax.swing.JScrollPane;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.DefaultTableModel;

import mpi.eudico.client.annotator.ElanLocale;
import mpi.eudico.client.annotator.commands.Command;
import mpi.eudico.client.annotator.commands.ELANCommandFactory;
import mpi.eudico.client.annotator.turnsandscenemode.commands.CancelEditAction;
import mpi.eudico.client.annotator.turnsandscenemode.commands.TaSActivateCellAction;
import mpi.eudico.client.annotator.turnsandscenemode.commands.TaSPostCutAction;
import mpi.eudico.client.annotator.turnsandscenemode.commands.TaSMergeAction;
import mpi.eudico.client.annotator.turnsandscenemode.commands.TaSMergeAction.MERGE_ROW;
import mpi.eudico.client.annotator.turnsandscenemode.commands.TaSPostPasteAction;
import mpi.eudico.client.annotator.turnsandscenemode.commands.TaSSplitAction;
import mpi.eudico.client.annotator.util.ClientLogger;
import mpi.eudico.client.annotator.viewer.AbstractViewer;
import mpi.eudico.client.mediacontrol.ControllerEvent;
import mpi.eudico.server.corpora.clom.Annotation;
import mpi.eudico.server.corpora.clom.Tier;
import mpi.eudico.server.corpora.clom.Transcription;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;
import mpi.eudico.server.corpora.clomimpl.type.LinguisticType;
import mpi.eudico.server.corpora.event.ACMEditEvent;
import mpi.eudico.server.corpora.event.ACMEditListener;

/**
 * A viewer that (currently) shows the annotations of a single tier and adds "virtual" annotations
 * for gaps, empty area where there are no annotations. These virtual annotations can be edited, 
 * in which case real annotations are created in the corresponding gaps. 
 * Real and virtual annotations can be split (with the Enter key) based on the current media position
 * and the current caret position in the text. 
 * The main unique characteristic of this viewer is that always the full time line, from 0 to the 
 * media duration is covered by either real or virtual annotations because these segments are the
 * main way to access the entire time line for segmentation and labeling. 
 *  
 * @author Han
 *
 */
@SuppressWarnings("serial")
public class TurnsAndSceneViewer extends AbstractViewer implements ACMEditListener, ListSelectionListener {
	private TranscriptionImpl transcription;
	private String transTierId = "transcript";
	private TierImpl curTier;
	private List<TaSAnno> segmentList;
	private JScrollPane scrollPane;
	private DefaultTableModel tableModel;
	private TaSTable annoTable;
	
	/** The enter keystroke without modifiers triggers the action on key press. The event is consumed then. */
	private KeyStroke SPLIT_KS = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0, false);
	private KeyStroke ESC_KS = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0, false);
	// hard coded now, should be taken from the action or the ShortcutUtil class
	private KeyStroke PLAY_SEL = KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, KeyEvent.SHIFT_DOWN_MASK, false);
	private KeyStroke MERGE_PREV_KS = KeyStroke.getKeyStroke(KeyEvent.VK_UP, 
			KeyEvent.CTRL_DOWN_MASK + KeyEvent.SHIFT_DOWN_MASK, false);// previous is UP visually in the table
	private KeyStroke MERGE_NEXT_KS = KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 
			KeyEvent.CTRL_DOWN_MASK + KeyEvent.SHIFT_DOWN_MASK, false);// next is DOWN in the table
	private KeyStroke ACTIVATE_PREV_KS = KeyStroke.getKeyStroke(KeyEvent.VK_UP, 
			KeyEvent.CTRL_DOWN_MASK, false);// previous is UP visually in the table
	private KeyStroke  ACTIVATE_NEXT_KS = KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 
			KeyEvent.CTRL_DOWN_MASK, false);// next is DOWN in the table
	private KeyStroke  ACTIVATE_CURRENT_TIME_KS = KeyStroke.getKeyStroke(KeyEvent.VK_SLASH, 
			KeyEvent.CTRL_DOWN_MASK, false);// next is DOWN in the table
	private TaSSplitAction splitAction;
	private TaSMergeAction mergePreviousAction;
	private TaSMergeAction mergeNextAction;
	private TaSActivateCellAction activatePreviousAction;
	private TaSActivateCellAction activateNextAction;
	private TaSCellPanel editorPanel;
	private TaSCellEditor cellEditor;
	private TaSCellRenderer cellRenderer;
	private JPopupMenu popupMenu;
	/* a field to cache the "current" row at the last media controller update */
	private int rowAtMediaTime = -1;
	private int scrollRowToLocation;
	private int primaryFontSize = 16;
	
	public TurnsAndSceneViewer() {
		super();
	}
	
	public TurnsAndSceneViewer(Transcription transcription) {
		this();
		this.transcription = (TranscriptionImpl) transcription;
		initComponents();
		
	}
		
	/**
	 * Temporary override.
	 * This needs to be done by the layout manager
	 */
	@Override
	public void addNotify() {
		super.addNotify();
		TaSActivateCellAction activateCurrentTimeAction = new TaSActivateCellAction(this, SwingConstants.CENTER);
		String activateCurrentCommand = "ACTIVATE_CURRENT";
		JRootPane rootPane = SwingUtilities.getRootPane(this);
		if (rootPane != null) {
			rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(ACTIVATE_CURRENT_TIME_KS, activateCurrentCommand);
			rootPane.getActionMap().put(activateCurrentCommand, activateCurrentTimeAction);
		}
	}

	/**
	 * Temporary override.
	 * This needs to be done by the layout manager
	 */
	@Override
	public void removeNotify() {
		super.removeNotify();
		String activateCurrentCommand = "ACTIVATE_CURRENT";
		JRootPane rootPane = SwingUtilities.getRootPane(this);
		if (rootPane != null) {
			rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).remove(ACTIVATE_CURRENT_TIME_KS);
			rootPane.getActionMap().remove(activateCurrentCommand);
		}
	}

	private void initComponents() {
		setLayout(new GridBagLayout());
		tableModel = new DefaultTableModel(0, 1);
		annoTable = new TaSTable(tableModel);
		annoTable.setTableHeader(null);
		
		// ties parts together
		editorPanel = new TaSCellPanel();
		// add actions, to be split off into another method?
		splitAction = new TaSSplitAction(this);
		String splitCommand = "SPLIT";
		editorPanel.getTextArea().getInputMap().put(SPLIT_KS, splitCommand);
		editorPanel.getTextArea().getActionMap().put(splitCommand, splitAction);
		Action psAction = ELANCommandFactory.getCommandAction(transcription, ELANCommandFactory.PLAY_SELECTION);
		editorPanel.getTextArea().getInputMap().put(PLAY_SEL, ELANCommandFactory.PLAY_SELECTION);
		editorPanel.getTextArea().getActionMap().put(ELANCommandFactory.PLAY_SELECTION, psAction);
		// bind Tab to play selection? or to go to next cell?
		KeyStroke TAB_KS = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, 0, false);
		String tabPS = "TAB_PLAY_SEL";
		editorPanel.getTextArea().getInputMap().put(TAB_KS, tabPS);
		editorPanel.getTextArea().getActionMap().put(tabPS, psAction);
		String escCommand = "ESCAPE";
		editorPanel.getTextArea().getInputMap().put(ESC_KS, escCommand);
		editorPanel.getTextArea().getActionMap().put(escCommand, new CancelEditAction(this));
		
		editorPanel.getTextArea().getActionMap().put("cutPost", new TaSPostCutAction(this));
		editorPanel.getTextArea().getActionMap().put("pastePost", new TaSPostPasteAction(this));
		
		mergePreviousAction = new TaSMergeAction(this, MERGE_ROW.PREVIOUS);
		mergePreviousAction.putValue(Action.NAME, ElanLocale.getString(ELANCommandFactory.MERGE_ANNOTATION_WB));
		mergePreviousAction.putValue(Action.ACCELERATOR_KEY, MERGE_PREV_KS);
		String mergPrevCommand = "MERGE_PREV";
		editorPanel.getTextArea().getInputMap().put(MERGE_PREV_KS, mergPrevCommand);
		editorPanel.getTextArea().getActionMap().put(mergPrevCommand, mergePreviousAction);
		
		mergeNextAction = new TaSMergeAction(this, MERGE_ROW.NEXT);
		mergeNextAction.putValue(Action.NAME, ElanLocale.getString(ELANCommandFactory.MERGE_ANNOTATION_WN));
		mergeNextAction.putValue(Action.ACCELERATOR_KEY, MERGE_NEXT_KS);
		String mergNextCommand = "MERGE_NEXT";
		editorPanel.getTextArea().getInputMap().put(MERGE_NEXT_KS, mergNextCommand);
		editorPanel.getTextArea().getActionMap().put(mergNextCommand, mergeNextAction);
		// move to next or previous cell/annotation
		activateNextAction = new TaSActivateCellAction(this, SwingConstants.NEXT);
		String activateNextCommand = "ACTIVATE_NEXT";
		editorPanel.getTextArea().getInputMap().put(ACTIVATE_NEXT_KS, activateNextCommand);
		editorPanel.getTextArea().getActionMap().put(activateNextCommand, activateNextAction);
		
		activatePreviousAction = new TaSActivateCellAction(this, SwingConstants.PREVIOUS);
		String activatePrevCommand = "ACTIVATE_PREV";
		editorPanel.getTextArea().getInputMap().put(ACTIVATE_PREV_KS, activatePrevCommand);
		editorPanel.getTextArea().getActionMap().put(activatePrevCommand, activatePreviousAction);
		
		popupMenu = new JPopupMenu();
		popupMenu.add(mergePreviousAction);
		popupMenu.add(mergeNextAction);
		
		// end add actions
		
		editorPanel.getTextArea().getDocument().addDocumentListener(new CellDocumentListener());
		cellEditor = new TaSCellEditor(this, editorPanel);
		cellRenderer = new TaSCellRenderer(this);
		
		annoTable.getColumnModel().getColumn(0).setCellRenderer(cellRenderer);
		annoTable.setShowGrid(false);
		annoTable.setRowSorter(null);
		annoTable.getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		annoTable.getColumnModel().getColumn(0).setCellEditor(cellEditor);
		
		scrollPane = new JScrollPane(annoTable);
		addComponentListener(new ViewerComponentListener());
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.weightx = 1.0;
		gbc.weighty = 1.0;
		this.add(scrollPane, gbc);
		//initTier();
		scrollRowToLocation = SwingConstants.TOP;
		annoTable.getSelectionModel().addListSelectionListener(this);
		//annoTable.addMouseListener(new TableMouseListener());
		annoTable.setComponentPopupMenu(popupMenu);
		editorPanel.setInheritsPopupMenu(true);
		editorPanel.getTextArea().setInheritsPopupMenu(true);
	}
	
	private void initTier() {
		if (transcription != null) {
			TierImpl t = transcription.getTierWithId(transTierId);
			if (t == null) {
				// create the tier
				if (transcription.getTiers().isEmpty()) {
                    LinguisticType type = new LinguisticType(transTierId);                     
                    transcription.addLinguisticType(type); 
                    
                    TierImpl tier = new TierImpl(transTierId, "", transcription, type);
                    transcription.addTier(tier);
                    tier.setDefaultLocale(null);
                    t = tier;
				} else {
					// take the first top level tier to work on
					List<TierImpl> topTiers = transcription.getTopTiers();
					if (!topTiers.isEmpty()) {
						t = topTiers.get(0);
					}
				}
			}
			if (t != null) {
				curTier = t;
				// create pseudo-annotations for gaps
				fillAnnotationList(t);
			}
		}
		
		recalculateAllRowHeights();
	}
	
	private void fillAnnotationList(Tier t) {
		if (t != null) {
			List<AbstractAnnotation> annos = (List<AbstractAnnotation>) t.getAnnotations();
			segmentList = new ArrayList<TaSAnno>();
			long timeSoFar = 0L;
			long mediaDuration = getMediaDuration();
			if (mediaDuration == 0) {// if the player is not initialized yet?
				mediaDuration = Long.MAX_VALUE;
			}
			
			if (annos.isEmpty()) {
					segmentList.add(new TaSAnno(timeSoFar, 
						mediaDuration));
			} else {
				for (int i = 0; i < annos.size(); i++) {
					AbstractAnnotation aa = annos.get(i);
					if (aa.getBeginTimeBoundary() > timeSoFar) {
						segmentList.add(new TaSAnno(timeSoFar, aa.getBeginTimeBoundary()));
					}
					segmentList.add(new TaSAnno(aa));
					timeSoFar = aa.getEndTimeBoundary();
					
					if (i == annos.size() - 1) {
						if (aa.getEndTimeBoundary() < mediaDuration) {
							segmentList.add(new TaSAnno(aa.getEndTimeBoundary(), mediaDuration));
						}
					}
				}
			}
			
			for (TaSAnno a2d : segmentList) {
				tableModel.addRow(new Object[]{a2d});
			}
		}
	}
	
	/**
	 * Inserts a new TaSAnno object encapsulating an annotation in the table and
	 * modifies or removes "gap" annotations that possibly exist in the time span of
	 * the new annotation. 
	 * 
	 * @param tasAnno a new encapsulated annotation
	 */
	private void insertNewAnnotation(TaSAnno tasAnno) {
		if (tasAnno.getAnnotation() == null) {
			// this is currently not expected to be the case
			ClientLogger.LOG.warning("Cannot insert an empty (gap) annotation.");
			return;
		}
		
		int[] overlRows = getOverlappingRows(tasAnno.getBeginTime(), tasAnno.getEndTime());
		
		if (overlRows[0] == -1 || overlRows[1] == -1) {
			ClientLogger.LOG.warning("Error inserting a new row, there should be at least one "
					+ "time-overlapping row already in the table");
			return;
		}
		
		if (overlRows[0] == overlRows[1]) {
			// should be an exact time match, should be a gap
			TaSAnno curRow = (TaSAnno) tableModel.getValueAt(overlRows[0], 0);
			if (curRow.getBeginTime() == tasAnno.getBeginTime() && 
					curRow.getEndTime() == tasAnno.getEndTime()) {
				tableModel.removeRow(overlRows[0]);
				tableModel.insertRow(overlRows[0], new Object[]{tasAnno});
				calculateRowHeight(overlRows[0]);
			} else if (curRow.getBeginTime() == tasAnno.getBeginTime()) {// && curRow.getEndTime() > tasAnno.getEndTime()
				// insert the new annotation, modify the existing annotation
				curRow.setBeginTime(tasAnno.getEndTime());
				tableModel.insertRow(overlRows[0], new Object[]{tasAnno});
				calculateRowHeight(overlRows[0]);
				calculateRowHeight(overlRows[0] + 1);
			}			
			else {
				ClientLogger.LOG.warning("Cannot replace one single row because the time values "
						+ "of the new annotation are not equal to the existing ones.");
			}
		} else {//error
			ClientLogger.LOG.warning("There are more than one overlapping segments on this tier "
					+ "cannot insert the new annotation");
		}
	}
	
	/**
	 * Returns an int array of size 2, indicating the first and the last index of overlapping
	 * rows. These indexes can be the same. If one of the indexes is -1 there is no overlap 
	 * (which would be an error).
	 * 
	 * @param bt
	 * @param et
	 * @return an array of size 2
	 */
	private int[] getOverlappingRows(long bt, long et) {
		// is there a better way of finding the rows than iterating the rows of the model?
		int startIndex = -1, endIndex = -1;
		for (int i = 0 ; i < tableModel.getRowCount(); i++) {
			TaSAnno rowAnno = (TaSAnno) tableModel.getValueAt(i, 0);
			if (rowAnno.getEndTime() <= bt) {
				continue;
			}
			if (rowAnno.getEndTime() > bt && rowAnno.getBeginTime() < et) {
				if (startIndex == -1) {
					startIndex = i;
				}
				if (i == tableModel.getRowCount() - 1) {
					endIndex = i;
				}
			}
			if (rowAnno.getBeginTime() >= et) {
				if (endIndex == -1) {
					endIndex = i - 1;
				}
				break;
			}
		}
		
		return new int[]{startIndex, endIndex};
	}
	
	/**
	 * Returns the row for the given time value.
	 * 
	 * @param time the time to look up the row for
	 * @return the row index or -1 if not found
	 */
	private int getRowAtTime(long time) {
		for (int i = 0 ; i < tableModel.getRowCount(); i++) {
			TaSAnno rowAnno = (TaSAnno) tableModel.getValueAt(i, 0);
			if (time >= rowAnno.getBeginTime() && time < rowAnno.getEndTime()) {
				return i;
			}
		}
		return -1;
	}
	
	/**
	 * Check whether the annotation is in the table and returns the row index, -1 if not in the table.
	 * 
	 * @param a the annotation
	 * @return the row index or -1
	 */
	private int getRowForAnnotation(Annotation a) {
		for (int i = 0 ; i < tableModel.getRowCount(); i++) {
			TaSAnno rowAnno = (TaSAnno) tableModel.getValueAt(i, 0);
			if (rowAnno.getAnnotation() == a) {
				return i;
			}
		}
		return -1;
	}
	
	/**
	 * When the times of an annotation changed, neighboring virtual annotations may have to be 
	 * updated or created.
	 *  
	 * @param annotation the annotation hat changed
	 */
	private void updateNeighborSegments(Annotation annotation) {
		if (annotation != null) {
			// find the row for the annotation
			int rowIndex = getRowForAnnotation(annotation);
			
			if (rowIndex == -1) {
				ClientLogger.LOG.warning("Error updating neighboring rows, the annotation was not found in the table");
				return;
			}
			
			// check before and after this annotation if there are gaps
			TaSAnno prevAnno = null;
			TaSAnno nextAnno = null;
			
			if (rowIndex > 0) {
				prevAnno = (TaSAnno) tableModel.getValueAt(rowIndex - 1, 0);
			}
			if (rowIndex < tableModel.getRowCount() - 1) {
				nextAnno = (TaSAnno) tableModel.getValueAt(rowIndex + 1, 0);
			}
			
			if (prevAnno == null) {
				if (annotation.getBeginTimeBoundary() > 0) {// unlikely this can happen
					tableModel.insertRow(0, new Object[]{new TaSAnno(0, 
							annotation.getBeginTimeBoundary())});
					calculateRowHeight(rowIndex);
					rowIndex++;
					//calculateRowHeight(rowIndex);// 1
				}
			} else {// compare with previous annotation
				if (annotation.getBeginTimeBoundary() > prevAnno.getEndTime()) {
					if (prevAnno.getAnnotation() != null) {
						// insert before
						tableModel.insertRow(rowIndex, new Object[]{new TaSAnno(prevAnno.getEndTime(), 
								annotation.getBeginTimeBoundary())});
						calculateRowHeight(rowIndex);
						rowIndex++;
					} else {
						prevAnno.setEndTime(annotation.getBeginTimeBoundary());
					}
				} else if (annotation.getBeginTimeBoundary() < prevAnno.getEndTime()) {// unlikely
					if (prevAnno.getAnnotation() == null) {
						prevAnno.setEndTime(annotation.getBeginTimeBoundary());
						// if the gap duration == 0, remove it
						if (prevAnno.getBeginTime() == prevAnno.getEndTime()) {
							tableModel.removeRow(rowIndex - 1);
						}
					} // the else case would be an error
				}
			}
			
			if (nextAnno == null) {// last annotation in the list
				if (annotation.getEndTimeBoundary() < getMediaDuration()) {
					// add row to the end
					tableModel.addRow(new Object[]{new TaSAnno(annotation.getEndTimeBoundary(), 
							getMediaDuration())});
					calculateRowHeight(tableModel.getRowCount() - 1);
				}
			} else {// compare with next annotation
				if (annotation.getEndTimeBoundary() < nextAnno.getBeginTime()) {
					if (nextAnno.getAnnotation() != null) {
						// insert after
						tableModel.insertRow(rowIndex + 1, new Object[]{new TaSAnno(annotation.getEndTimeBoundary(), 
								nextAnno.getBeginTime())});
						calculateRowHeight(rowIndex + 1); 
					} else {
						nextAnno.setBeginTime(annotation.getEndTimeBoundary());
					}
				} else if (annotation.getEndTimeBoundary() > nextAnno.getBeginTime()) {
					if (nextAnno.getAnnotation() == null) {
						nextAnno.setBeginTime(annotation.getEndTimeBoundary());
						// if the gap duration == 0, remove it
						if (nextAnno.getBeginTime() == nextAnno.getEndTime()) {
							tableModel.removeRow(rowIndex + 1);
						}
					} else { 
						// this can happen when annotations are merged. The remove annotation ACMEvent in that case 
						// is too generic to identify which annotation has been removed
						if (nextAnno.getBeginTime() == nextAnno.getEndTime()) {
							tableModel.removeRow(rowIndex + 1);
						}
					}
				}
			}
		}
	}
	
	private void updateRow(Annotation annotation) {
		if (annotation != null) {
			for (int i = 0; i < tableModel.getRowCount(); i++) {
				TaSAnno rowAnno = (TaSAnno) tableModel.getValueAt(i, 0);
				if (rowAnno.getAnnotation() == annotation) {
					calculateRowHeight(i);
					break;
				}
			}
		}
	}
	
	/**
	 * When an annotation has been removed from the tier, remove the corresponding row from the 
	 * view and replace it by an empty virtual annotation.
	 * 
	 * @param remAnn the removed annotation
	 */
	private void annotationRemoved(Annotation remAnn) {
		if (remAnn != null) {// superfluous?
			for (int i = 0; i < tableModel.getRowCount(); i++) {
				TaSAnno rowAnno = (TaSAnno) tableModel.getValueAt(i, 0);
				if (rowAnno.getAnnotation() == remAnn) {
					//tableModel.removeRow(i);
					
					// check before and after this annotation if there are gaps
					boolean gapBefore = false;
					boolean gapAfter = false;
					TaSAnno prevAnno = null;
					TaSAnno nextAnno = null;
					
					if (i > 0) {
						prevAnno = (TaSAnno) tableModel.getValueAt(i - 1, 0);
						gapBefore = (prevAnno.getAnnotation() == null); 
					}
					if (i < tableModel.getRowCount() - 1) {
						nextAnno = (TaSAnno) tableModel.getValueAt(i + 1, 0);
						gapAfter = (nextAnno.getAnnotation() == null);
					}
					
					if (gapBefore && gapAfter) {
						// remove row at i, remove row at i + 1, update gap row at i - 1
						tableModel.removeRow(i + 1);// first delete the next
						tableModel.removeRow(i);
						prevAnno.setEndTime(nextAnno.getEndTime());
						calculateRowHeight(i - 1);
					} else if (gapBefore) {
						tableModel.removeRow(i);
						prevAnno.setEndTime(remAnn.getEndTimeBoundary());
						calculateRowHeight(i - 1);
					} else if (gapAfter) {
						tableModel.removeRow(i);
						nextAnno.setBeginTime(remAnn.getBeginTimeBoundary());
						calculateRowHeight(i);// nextAnno now in row i
					} else { // no adjacent gaps
						tableModel.removeRow(i);
						tableModel.insertRow(i, new Object[]{new TaSAnno(remAnn.getBeginTimeBoundary(), 
								remAnn.getEndTimeBoundary())});
						calculateRowHeight(i);
					}
					
					break;
				}
			}
		}
	}
	
	/**
	 * Calculates (and sets) the height of the specified row. Currently relies on
	 * the preferred size of the rendering component.
	 * 
	 * @param row the row
	 * @return the preferred height of the row
	 */
	public int calculateRowHeight(int row) {
		if (row < 0 || row >= annoTable.getRowCount()) {
			throw new IllegalArgumentException("There is no row with index: " + row);
		}
		
		TaSCellPanel renderComp = (TaSCellPanel) cellRenderer.getTableCellRendererComponent(annoTable, 
				tableModel.getValueAt(row, 0), false, false, row, 0);
		int prefHeight = renderComp.getPreferredSize().height;
		annoTable.setRowHeight(row, prefHeight);// or do not set it here?
		if (row == annoTable.getRowCount() - 1) {
			annoTable.setRowHeight(row, 2 * prefHeight);// extra space for the last row
		}
		
		return prefHeight;
	}
	
	public void recalculateAllRowHeights() {
		for (int i = 0; i < annoTable.getRowCount(); i++) {
			calculateRowHeight(i);
		}
		
		annoTable.revalidate();
	}
	
	/**
	 * Called by the special split/commit action that has been added to the text area.
	 * 
	 * 1) If there was not an annotation yet in the editing row, an annotation will be created
	 * with the text in the cell and with the media time as end time. If the caret is not at the
	 * end of the text 2 annotations will be created in the current segment.
	 * 
	 * 2) If there was already an annotation in the row it will be split at the text position and 
	 * at the current media position. 
	 * Some trade-off decisions have to be made when the media time is not within the current
	 * segment (if that can happen at all).
	 */
	public void splitEventOccurred() {
		//System.out.println("Split Event Occurred");
		// get the editor panel, get access to the text area, get text and caret location
		// get current time, split if possible, warn if not
//		System.out.println("Split: Text: " + editorPanel.getTextArea().getText());
//		System.out.println("Split: Text Length: " + editorPanel.getTextArea().getText().length());
//		System.out.println("Split: Caret: " + editorPanel.getTextArea().getCaretPosition());
//		System.out.println("Split: Time: " + getMediaTime());
		String curText = editorPanel.getTextArea().getText();
		int caretPos = editorPanel.getTextArea().getCaretPosition();
		long curMediaTime = getMediaTime();
		int editingRow = annoTable.getEditingRow();
		// add a test for -1, should not happen
		TaSAnno curAnno = (TaSAnno) annoTable.getValueAt(editingRow, 0);
		// one or both of the next calls?
		editorPanel.getTextArea().transferFocusUpCycle();
		cellEditor.stopCellEditing();

		// now create editing action(s) 
		// if the media time is not within the current segment, no splitting can be done
		boolean shouldSplit = curMediaTime > curAnno.getBeginTime() && curMediaTime < curAnno.getEndTime();
		if (curAnno.getAnnotation() != null) {
			// edit an existing annotation and possibly create a new one
			// only allow splitting of an annotation when it has no dependent annotations in this mode
			boolean annHasDepending = !((AbstractAnnotation) curAnno.getAnnotation()).getParentListeners().isEmpty();
			
			if (!shouldSplit || (shouldSplit && annHasDepending)) {// do the same in both cases, modify the text
				if (shouldSplit && annHasDepending) {
					ClientLogger.LOG.warning("Splitting of annotations that have depending annotations is not supported.");
					// show a message dialog
				}
				// no way to split, ignore the caret position, check if the text changed 
				if (curText != null && !curText.equals(curAnno.getAnnotation().getValue())) {
					Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MODIFY_ANNOTATION);
					com.execute(curAnno.getAnnotation(), new Object[]{curAnno.getAnnotation().getValue(), curText});
					// wait for the ACM event to update the TaSAnno object in the table
				}
			} else /*if (shouldSplit && !annHasDepending)*/ {	
				Command splitCom = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.SPLIT_ANNOTATION);
				Object[] args = new Object[4];
				args[0] = curAnno.getAnnotation();
				args[1] = curMediaTime;
				args[2] = curText;
				args[3] = caretPos;
				splitCom.execute(transcription, args);
			}
		} else { // a gap
			// create one or two new annotations
			long splitTime = shouldSplit ? curMediaTime : -1;
			Command twoCom = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.NEW_ANNOTATIONS_IN_GAP);
			twoCom.execute(curTier, new Object[]{curAnno.getBeginTime(), curAnno.getEndTime(), splitTime, curText, caretPos});
		}
		
		annoTable.editCellAt(editingRow + 1, 0);
		cellEditor.startEditing();
	}
	
	/**
	 * Called by the merge with next action fore this mode.
	 * 
	 * This implementation is probably best moved to the action itself, depending on needs to be made accessible to the 
	 * action then (what is now only accessible in the class).
	 */
	public void mergeWithNextEventOccurred() {
		int editingRow = annoTable.getEditingRow();
		if (editingRow == -1) {// should not occur
			ClientLogger.LOG.info("Cannot merge with the next cell: there is no cell selected.");
			return;
		}
		if (editingRow == annoTable.getRowCount() - 1) {
			ClientLogger.LOG.info("Cannot merge with the next cell: the active cell is the last cell.");
			return;
		}
		//String curText = editorPanel.getTextArea().getText();//?? do something with the text which might have been changed?
		TaSAnno curAnno = (TaSAnno) annoTable.getValueAt(editingRow, 0);
		TaSAnno nextAnno = (TaSAnno) annoTable.getValueAt(editingRow + 1, 0);
		// one or both of the next calls?
		editorPanel.getTextArea().transferFocusUpCycle();
		cellEditor.stopCellEditing();
		
		if (curAnno.getAnnotation() != null) {
			if (nextAnno.getAnnotation() != null) {
				// merge two annotations, reuse existing command
				// only allow merging if there are no dependent annotations for now (mainly because of to generic ACMEditEvent)
				boolean ann1HasDepending = !((AbstractAnnotation) curAnno.getAnnotation()).getParentListeners().isEmpty();
				boolean ann2HasDepending = !((AbstractAnnotation) nextAnno.getAnnotation()).getParentListeners().isEmpty();
				if (!ann1HasDepending && !ann2HasDepending) {
					Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MERGE_ANNOTATION_WN);
					com.execute(transcription, new Object[]{curAnno.getAnnotation(), Boolean.TRUE});
				} else {
					// show message
				}
			} else {
				// merge with next gap == modify annotation time to include the next gap's time interval
				Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MODIFY_ANNOTATION_TIME);
				com.execute(curAnno.getAnnotation(), new Object[]{curAnno.getBeginTime(), nextAnno.getEndTime()});
			}
		} else { // the active cell is a gap, the next cell must be an annotation 
			if (nextAnno.getAnnotation() != null) {
				// modify the next annotation's time by adding the time interval of this gap
				Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MODIFY_ANNOTATION_TIME);
				com.execute(nextAnno.getAnnotation(), new Object[]{curAnno.getBeginTime(), nextAnno.getEndTime()});
			} else {
				ClientLogger.LOG.warning("Unexpected state of the sequence of annotations: cannot merge a gap with another gap");
			}
		}
		
		annoTable.editCellAt(editingRow, 0);
		cellEditor.startEditing();
	}

	
	/**
	 * Called by the merge with previous action fore this mode.
	 * 
	 * This implementation is probably best moved to the action itself, depending on needs to be made accessible to the 
	 * action then (what is now only accessible in the class).
	 */
	public void mergeWithPreviousEventOccurred() {
		int editingRow = annoTable.getEditingRow();
		if (editingRow == -1) {// should not occur
			ClientLogger.LOG.info("Cannot merge with the previous cell: there is no cell selected.");
			return;
		}
		if (editingRow == 0) {
			ClientLogger.LOG.info("Cannot merge with the previous cell: the active cell is the first cell.");
			return;
		}
		//String curText = editorPanel.getTextArea().getText();//?? do something with the text which might have been changed?
		TaSAnno curAnno = (TaSAnno) annoTable.getValueAt(editingRow, 0);
		TaSAnno prevAnno = (TaSAnno) annoTable.getValueAt(editingRow - 1, 0);
		// one or both of the next calls?
		editorPanel.getTextArea().transferFocusUpCycle();
		cellEditor.stopCellEditing();
		
		if (curAnno.getAnnotation() != null) {
			if (prevAnno.getAnnotation() != null) {
				// merge two annotations, reuse existing command
				// only allow merging if there are no dependent annotations for now (mainly because of to generic ACMEditEvent)
				boolean ann1HasDepending = !((AbstractAnnotation) curAnno.getAnnotation()).getParentListeners().isEmpty();
				boolean ann2HasDepending = !((AbstractAnnotation) prevAnno.getAnnotation()).getParentListeners().isEmpty();
				if (!ann1HasDepending && !ann2HasDepending) {
					Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MERGE_ANNOTATION_WB);
					com.execute(transcription, new Object[]{curAnno.getAnnotation(), Boolean.FALSE});
				} else {
					// show message
				}
			} else {
				// merge with previous gap == modify annotation time to include the previous gap's time interval
				Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MODIFY_ANNOTATION_TIME);
				com.execute(curAnno.getAnnotation(), new Object[]{prevAnno.getBeginTime(), curAnno.getEndTime()});
			}
		} else { // the active cell is a gap, the previous cell must be an annotation 
			if (prevAnno.getAnnotation() != null) {
				// modify the previous annotation's time by adding the time interval of this gap
				Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MODIFY_ANNOTATION_TIME);
				com.execute(prevAnno.getAnnotation(), new Object[]{prevAnno.getBeginTime(), curAnno.getEndTime()});
			} else {
				ClientLogger.LOG.warning("Unexpected state of the sequence of annotations: cannot merge a gap with another gap");
			}
		}
		
		annoTable.editCellAt(editingRow - 1, 0);
		cellEditor.startEditing();
	}
	
	/**
	 * Commits the text of the current cell (in case this is already an existing annotation) after 
	 * a "Cut" action occurred. Otherwise the text would still be in the annotation if the cell 
	 * is de-activated without e.g. hitting Enter. 
	 */
	public void cutActionOccurred() {
		int editingRow = annoTable.getEditingRow();
		if (editingRow == -1) {// should not occur
			ClientLogger.LOG.info("A cut action occurred but it is unknown in which cell.");
			return;
		}
		TaSAnno curAnno = (TaSAnno) annoTable.getValueAt(editingRow, 0);
		if (curAnno.getAnnotation() != null) {
			// commit the current text
			String curText = editorPanel.getTextArea().getText();
			if (curText != null && !curText.equals(curAnno.getAnnotation().getValue())) {
				Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MODIFY_ANNOTATION);
				com.execute(curAnno.getAnnotation(), new Object[]{curAnno.getAnnotation().getValue(), curText});
				// wait for the ACM event to update the TaSAnno object in the table, 
				// this deactivates the cell, not sure if this is always the desired behavior
			}
		}
	}
	
	/**
	 * Commits the text of the current cell after a "Paste" action. Commits to the current annotation
	 * or creates a new annotation in case of a gap. 
	 */
	public void pasteActionOccurred() {
		int editingRow = annoTable.getEditingRow();
		if (editingRow == -1) {// should not occur
			ClientLogger.LOG.info("A paste action occurred but it is unknown in which cell.");
			return;
		}
		TaSAnno curAnno = (TaSAnno) annoTable.getValueAt(editingRow, 0);
		String curText = editorPanel.getTextArea().getText();
		if (curAnno.getAnnotation() != null) {
			// commit the current text
			if (curText != null && !curText.equals(curAnno.getAnnotation().getValue())) {
				Command com = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.MODIFY_ANNOTATION);
				com.execute(curAnno.getAnnotation(), new Object[]{curAnno.getAnnotation().getValue(), curText});
				// wait for the ACM event to update the TaSAnno object in the table, 
				// this deactivates the cell, not sure if this is always the desired behavior
			}
		} else {// create an annotation
			// create one or two new annotations
			int caretPos = editorPanel.getTextArea().getCaretPosition();
			long splitTime = -1;
			Command twoCom = ELANCommandFactory.createCommand(transcription, ELANCommandFactory.NEW_ANNOTATIONS_IN_GAP);
			twoCom.execute(curTier, new Object[]{curAnno.getBeginTime(), curAnno.getEndTime(), splitTime, curText, caretPos});
			// wait for the ACM event to update the TaSAnno object in the table, 
			// this deactivates the cell, not sure if this is always the desired behavior
		}
	}
	
	
	/**
	 * A call to activate the previous or next segment/cell in the view.
	 * This method tries to place the caret at the end of the text of the segment that became active
	 * (As with the other responses to (keyboard) actions this implementation should be reconsidered.)
	 * 
	 * @param direction up or down, one of SwingConstants PREVIOUS or NEXT, 
	 * or cell at current time in case of SwingConstants CENTER
	 */
	public void activateSegment(int direction) {
		int editingRow = annoTable.getEditingRow();
		
		int nextEditingRow = 0;
		switch (direction) {
		case SwingConstants.CENTER:
			nextEditingRow = getRowAtTime(getMediaTime());
			
			if (nextEditingRow == -1) {
				return;
			}
			break;
		case SwingConstants.PREVIOUS:
			if (editingRow == 0) {
				return;
			}
			nextEditingRow = editingRow - 1;
			
			break;
		case SwingConstants.NEXT:
			if (editingRow == annoTable.getRowCount() - 1) {
				return;
			}
			nextEditingRow = editingRow + 1;
			break;
		default:
			// nothing	
		}
		
		cellEditor.stopCellEditing();
		annoTable.getSelectionModel().setSelectionInterval(nextEditingRow, nextEditingRow);
		annoTable.editCellAt(nextEditingRow, 0);
		editorPanel.startEditing();
		editorPanel.getTextArea().setCaretPosition(editorPanel.getTextArea().getText().length());
	}

	/**
	 * Informs the table to cancel editing.
	 */
	public void cancelEditing() {
		if (annoTable.getEditingRow() > -1) {
			cellEditor.cancelCellEditing();
		}
	}
	
	@Override
	public void controllerUpdate(ControllerEvent event) {
		// TODO take the current, cached segment and check if the crosshair
		// is still in the same segment. Otherwise find the segment we are now in and highlight it, repaint
		long mediaTime = getMediaTime();
		int editingRow = annoTable.getEditingRow();
		
		if (editingRow > -1) {
			cellEditor.repaint();
			// this cell probably is already in view, but check
			//...
		}
		int medRow = getRowAtTime(mediaTime);
		if(medRow > -1 && medRow != rowAtMediaTime) {
			scrollToRow(medRow);
			rowAtMediaTime = medRow;
		}
		
		repaint();
	}

	@Override
	public void updateSelection() {
		repaint();
	}

	@Override
	public void updateActiveAnnotation() {
		if (annoTable != null) {
			
		}
	}

	@Override
	public void updateLocale() {
		// TODO implement when needed

	}

	@Override
	public void preferencesChanged() {
		// TODO check font size, display settings etc.
//		if (annoTable != null && annoTable.getFont() != null) {
//			annoTable.setFont(annoTable.getFont().deriveFont((float) primaryFontSize));
//		}
		initTier();
	}

	/**
	 * In response to edit actions the table possibly has to be updated in several ways. The height of rows might 
	 * need to be recalculated, new gaps might need to be filled with virtual annotations or virtual annotations
	 * might need to be modified or removed.
	 */
	@Override
	public void ACMEdited(ACMEditEvent e) {
		//System.out.println("ACM Event: " + e.getOperation() + " M: " + e.getModification() + " I: " + e.getInvalidatedObject());
		cellEditor.cancelCellEditing();// more elegant way to stop editing before modifying the table?
		// handle (for now) new annotation, modify annotation, delete annotation, delete tier? (if it is the current tier)
		switch (e.getOperation()) {
		case ACMEditEvent.ADD_ANNOTATION_HERE:
			// add a new TaSAnno object possibly in place of an empty placeholder and remove or update empty placeholder
            if (e.getInvalidatedObject() instanceof TierImpl &&
                    e.getModification() instanceof Annotation) {
            	if (e.getInvalidatedObject() == curTier) {
            		insertNewAnnotation(new TaSAnno( (Annotation) e.getModification()));
            	}
            }
			break;
		
		case ACMEditEvent.CHANGE_ANNOTATION_TIME:
			// find the row, check neighboring annotations/cells to see if they have to be updated
            if (e.getModification() instanceof Annotation) {
            	Annotation modAnno = (Annotation) e.getModification();
            	if (modAnno.getTier() == curTier) {
            		updateNeighborSegments(modAnno);
            	}
            }
			break;
		
		case ACMEditEvent.CHANGE_ANNOTATION_VALUE:
			// find the row, calculate the height, re-validate the table
            if (e.getInvalidatedObject() instanceof Annotation) {
            	updateRow((Annotation) e.getInvalidatedObject());
            }
			break;
			
		case ACMEditEvent.CHANGE_ANNOTATIONS:
			// it would be best if this can be avoided
			break;
			
		case ACMEditEvent.REMOVE_ANNOTATION:
			// find the row, remove it, update/insert placeholder
			if (e.getModification() instanceof Annotation) {
				Annotation remAnn = (Annotation) e.getModification();
				// check if the tier is in the table, currently only one tier
				if (remAnn.getTier() == curTier) {
					annotationRemoved(remAnn);
				}
			}
			break;
		case ACMEditEvent.REMOVE_TIER:
			// check it the removed tier is (one of) the current tier(s); if so remove it from the table
			if (e.getModification() instanceof TierImpl) {
				if (e.getModification() == curTier) {
					// clear the table
					tableModel.setRowCount(0);
					this.revalidate();// necessary?
				}
			}
			break;
		case ACMEditEvent.CHANGE_TIER:
			// if the changed tier is (one of) the current tier(s) update tier and participant labels, if necessary 
		default:
			break;
		}
	}
	
	/**
	 * List Selection event handling
	 */
	@Override
	public void valueChanged(ListSelectionEvent e) {
		//System.out.println("List Selection Event: F: " + e.getFirstIndex() + " L: " + e.getLastIndex());
		if (!e.getValueIsAdjusting()) {		
			int selRow = annoTable.getSelectedRow();
			//System.out.println("Selected Row: " + selRow);
			if (selRow > -1) {
				TaSAnno selAnno = (TaSAnno) annoTable.getValueAt(selRow, 0);
				setSelection(selAnno.getBeginTime(), selAnno.getEndTime());
				setMediaTime(selAnno.getBeginTime());
				//setActiveAnnotation(selAnno.getAnnotation()); // can be null
				annoTable.repaint();
			}
		}		
	}

	@Override
	public void setBounds(int x, int y, int width, int height) {
		int curWidth = getWidth();
		super.setBounds(x, y, width, height);
		if (curWidth != width) {
			recalculateAllRowHeights();
		}
		this.revalidate();
	}
	
	/**
	 * Sets the main font size for the viewer, i.e. the table
	 * 
	 * @param fontSize the new size of the font
	 */
	public void setFontSize(int fontSize) {
		if (this.primaryFontSize != fontSize) {
			this.primaryFontSize = fontSize;
			if (annoTable != null) {
				annoTable.setFont(annoTable.getFont().deriveFont((float) this.primaryFontSize));
				if (annoTable.isEditing()) {
					((TaSCellPanel) cellEditor.getComponent()).setFont(annoTable.getFont());
				}
				recalculateAllRowHeights();
			}
		}
	}
	
	/**
	 * Returns the main font size for this viewer (the size of the text area in the table cells) 
	 * @return the primary font size
	 */
	public int getFontSize() {
		return primaryFontSize;
	}
	
	/**
	 * Checks if the given row is visible, scrolls the table if necessary.
	 * Based on scrollIfNeeded in AbstractGridViewer.
	 * 
	 * @param row the row to be made visible
	 */
	private void scrollToRow(int row) {
		if (row > -1) {
			synchronized (annoTable) {
				//System.out.println("Scroll 0: ViewSize: " + scrollPane.getViewport().getViewSize());
				Dimension viewSize = scrollPane.getViewport().getViewSize();
				Rectangle viewRect = scrollPane.getViewport().getViewRect();
				Rectangle cellRect = annoTable.getCellRect(row, 0, true);
				//System.out.println("Scroll 1: View: " + viewRect + " Cell: " + cellRect);
				
				if (!viewRect.contains(cellRect)) {
					// reuse the cellRect Rectangle to calculate the new rectangle that should be made visible 
					switch (scrollRowToLocation) {
					case SwingConstants.TOP:
						cellRect.height = (viewRect.height - cellRect.height);
						
						break;
					case SwingConstants.CENTER:
						cellRect.height += (viewRect.height - cellRect.height) / 2;
						
						break;
					case SwingConstants.BOTTOM:
						cellRect.height += 10;
						
						break;
						
						default:
							
					}
					
					if (cellRect.y + cellRect.height > viewSize.height) {
						//System.out.println("Scroll: Adjust Cell Height1: " + cellRect);
						cellRect.height = viewSize.height - cellRect.y;
						//System.out.println("Scroll: Adjust Cell Height2: " + cellRect);
					}
					//cellRect.height = (viewRect.height - cellRect.height) / 2;//??
					//System.out.println("Scroll 2: View: " + viewRect + " Cell: " + cellRect);
					annoTable.scrollRectToVisible(cellRect);
				}
			}	
		}
	}
	
	/**
	 * Triggers an update of the row heights after e.g. resizing of the viewer.
	 */
	private class ViewerComponentListener implements ComponentListener {

		@Override
		public void componentResized(ComponentEvent e) {
			recalculateAllRowHeights();
//			System.out.println("Resized...");
		}

		@Override
		public void componentMoved(ComponentEvent e) {
			
		}

		@Override
		public void componentShown(ComponentEvent e) {
			recalculateAllRowHeights();
//			System.out.println("Shown...");
		}

		@Override
		public void componentHidden(ComponentEvent e) {
			
		}
	}
	
	/**
	 * A listener to document changes so that the height of the editing row in the table can be
	 * recalculated. 
	 */
	private class CellDocumentListener implements DocumentListener {
		@Override
		public void insertUpdate(DocumentEvent e) {
			//System.out.println("Inserted into Document at row: " + annoTable.getEditingRow());
			//System.out.println("Length of insertion: " + e.getLength());
			// could also try to get the location of the last character/line in the view
			/*
			try {
				Rectangle lastRect = editorPanel.getTextArea().modelToView(editorPanel.getTextArea().getText().length());
				curAreaHeight = lastRect.y;
				System.out.println("End rect: " + lastRect.y + " - " + lastRect.height);
			} catch (BadLocationException ble) {
				System.out.println("Bad: " + ble.getMessage());
			}
			*/
			// when a longer text has been pasted, the preferred size is not immediately updated, for some reason
			if (e.getLength() > 1) {
				editorPanel.getTextArea().revalidate();
				editorPanel.getTextArea();
				editorPanel.revalidate();
			}
			//curEditingRow = annoTable.getEditingRow();
			checkRowHeight(annoTable.getEditingRow());
		}

		@Override
		public void removeUpdate(DocumentEvent e) {
			//System.out.println("Removed from Document at row: " + annoTable.getEditingRow());
			checkRowHeight(annoTable.getEditingRow());
		}

		@Override
		public void changedUpdate(DocumentEvent e) {
			// stub maybe check if font size is changed etc?
			//System.out.println("Change update from Document at row: " + annoTable.getEditingRow());
		}
		
		private void checkRowHeight(int row) {
			if (row > -1) {		
				int curHeight = editorPanel.getTextArea().getHeight();
				int curPrefHeight = editorPanel.getTextArea().getPreferredSize().height;
				//System.out.println("Cur H: " + curHeight + "  Pref H: " + curPrefHeight);
				if (curHeight != curPrefHeight) {
					if ( !(row == annoTable.getRowCount() - 1 && curHeight > curPrefHeight)) {
						annoTable.setRowHeight(row, annoTable.getRowHeight(row) + (curPrefHeight - curHeight));
						annoTable.revalidate();
						editorPanel.revalidate();
					}
				}
			}
		}
	}
	
	private class TableMouseListener implements MouseListener {

		@Override
		public void mouseClicked(MouseEvent e) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void mousePressed(MouseEvent e) {
			if (e.getSource() == annoTable) {
		        if ((SwingUtilities.isRightMouseButton(e) && (e.getButton() == MouseEvent.BUTTON1 ^ e.isMetaDown())) 
		                || e.isPopupTrigger()) {
					popupMenu.show(annoTable, e.getX(), e.getY());
				}
			}
		}

		@Override
		public void mouseReleased(MouseEvent e) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void mouseEntered(MouseEvent e) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void mouseExited(MouseEvent e) {
			// TODO Auto-generated method stub
			
		}
		
	}
}
