package mpi.eudico.client.annotator.util;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import mpi.eudico.client.util.TranscriptionECVLoader;
import mpi.eudico.server.corpora.clom.ExternalReference;
import mpi.eudico.server.corpora.clom.TranscriptionStore;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;
import mpi.eudico.server.corpora.clomimpl.dobes.ACMTranscriptionStore;
import mpi.eudico.server.corpora.clomimpl.type.LinguisticType;
import mpi.eudico.util.CVEntry;
import mpi.eudico.util.ControlledVocabulary;
import mpi.eudico.util.ExternalCV;

/**
 * A class/script that updates all transcriptions in a specified folder by 
 * checking external controlled vocabularies and updating, if needed, annotations
 * that are linked to a CV entry. The files can either be overwritten or
 * can be saved in a different folder. There is support for specifying which
 * language to use from the CV; this overrides the normal checks on the tier's
 * language and on stored language preferences.
 * TODO: this might be useful for internal CV's as well?
 */
public class CorpusECVUpdater {
	private File inputFolder;
	private String inputFolderString;
	private File outputFolder;
	private String outputFolderString;
	/** select the entry variant of this language from the CV, 
	 *  the value should correspond to the LANG_REF attribute 
	 *  in most cases a 3-letter code */
	private String forcedLanguage;
	
	/** a map of ecv url's to lists of (loaded) ECV's */
	private HashMap<String, List<ExternalCV>> urlMap;

	/**
	 * Constructor.
	 */
	private CorpusECVUpdater() {
	}

	/**
	 * For command-line usage:<br> 
	 * >java ... CorpusECVUpdater [-L lng] inputfolder [outputfolder]
	 * 
	 * Starts the updating process if at least an input folder is 
	 * provided. 
	 * 
	 * @param args the arguments should include the input folder, 
	 * can include a -L language parameter and an output folder
	 */
	public static void main(String[] args) {
		if (args.length == 0) {
			System.out.println("No folder containing eaf files specified");
			System.out.println("Usage: java (...) CorpusECVUpdater [-L lng] inputfolder [outputfolder]");
			System.exit(1);
		}
		String fLang = null;
		String[] folders = new String[2];
		if (args.length >= 3) {
			if (args[0].toLowerCase().equals("-l")) {
				fLang = args[1];
				System.out.println("Language to apply is: " + fLang);
				folders[0] = args[2];
				if (args.length > 3) {
					folders[1] = args[3];
				}
			} else {// simply take the first 2 as input & output folder 
				folders[0] = args[0];
				folders[1] = args[1];
			}
		} else {// args.length = 1 or 2
			folders[0] = args[0];
			if (args.length > 1) {
				folders[1] = args[1];
			}
		}
		System.out.println("Input folder is: " + folders[0]);
		if (folders[1] != null) {
			System.out.println("Output folder is: " + folders[1]);
		} else {
			System.out.println("No output folder specified, overwriting files");
		}
		// other tests?
		
		CorpusECVUpdater cECV = new CorpusECVUpdater();
		int success = cECV.updateFiles(folders, fLang);
		System.out.println("Exit Program: " + success);
		System.exit(success);
	}

	/**
	 * Updates all .eaf files in the specified input folder (recursively).
	 * 
	 * @param folders one input folders and possibly an output folder
	 * @param forceLang the language to select from the CV
	 * 
	 * @return a non-zero exit value in case there is an error, 0 otherwise
	 */
	private int updateFiles(String[] folders, String forceLang) {
		forcedLanguage = forceLang;
		try {
			inputFolder = new File(folders[0]);
			
			if (!inputFolder.exists() || !inputFolder.isDirectory()) {
				System.out.println("The specified path is not a folder containing eaf files: " + folders[0]);
				return 2;
			}
			inputFolderString = inputFolder.getAbsolutePath().replace('\\', '/');
			
			outputFolder = null;
			if (folders.length > 1 && folders[1] != null) {
				outputFolder = new File(folders[1]);
				if (!outputFolder.exists()){
					boolean created = outputFolder.mkdirs();
					if (!created) {
						System.out.println("Could not create the output folder: " + folders[1]);
						return 4;
					}
				}
				outputFolderString = outputFolder.getAbsolutePath().replace('\\', '/');
			}
			
			processFile1(inputFolder);
		} catch (Throwable t) {
			System.out.println("Error while processing files: " + t.getMessage());
			t.printStackTrace();
		}
		
		
		return 0;
	}
	
	/**
	 * A method that checks if the File object represents a folder or a file
	 * and then delegates to an appropriate method.
	 * 
	 * @param file a folder or a file
	 */
	private void processFile1(File file) {
		if (file.isDirectory()) {
			File[] subList = file.listFiles();
			for (File f : subList) {
				processFile1(f);
			}
		} else {
			processFile2(file);
		}
	}
	
	/**
	 * Processes the specified file. If it is not an eaf file it is 
	 * currently ignored.
	 * 
	 * @param file the file to update, folders are ignored
	 */
	private void processFile2(File file) {
		if (file.isDirectory()) {
			System.out.println("Error: method called with folder as argument");
		} else {
			try {
				if (file.getName().toLowerCase().endsWith(".eaf")) {
					TranscriptionImpl t = new TranscriptionImpl(file.getAbsolutePath());
					t.setUnchanged();
					
					if (t.getControlledVocabularies().size() > 0) {
						boolean allCVSReplaced = true;
						if (urlMap != null) {
							allCVSReplaced = updateAlreadyLoadedCVS(urlMap, t);
						} 
						if (urlMap == null || !allCVSReplaced) {
							TranscriptionECVLoader tl = new TranscriptionECVLoader();
							// this can set the changed flag, it can update annotations
							tl.loadExternalCVs(t, null);
							// TODO change the following double processing
							// to make sure that the specified language is enforced, do it again
							if (forcedLanguage != null) {
								updateAnnotations(t);
							}
							urlMap = new HashMap<String, List<ExternalCV>>(tl.getLoadedECVMap());
							Iterator<List<ExternalCV>> ecvIter = urlMap.values().iterator();
							while (ecvIter.hasNext()) {
								List<ExternalCV> ecvList = urlMap.get(ecvIter.next());
								if (ecvList != null) {
									for (ExternalCV ecv : ecvList) {
										if (ecv != null) {
											ecv.removeACMEditableObject();
										}
									}
								}
							}
						}
					}
					
					if (outputFolder == null) {
						if (t.isChanged()) {
							System.out.println("Updated file: " + file.getName());
							ACMTranscriptionStore.getCurrentTranscriptionStore().
							storeTranscription(t, null, null, TranscriptionStore.EAF);
						}
					} else {
						// calculate the new location
						String nextPath = file.getAbsolutePath().replace('\\', '/');
						nextPath = nextPath.replace(inputFolderString, outputFolderString);
						
						String outputSubFolder = nextPath.substring(0, nextPath.lastIndexOf('/'));
						if (!outputSubFolder.equals(outputFolderString)) {
							File osFile = new File(outputSubFolder);
							if (!osFile.exists()) {
								osFile.mkdirs();
							}
						}
						
						ACMTranscriptionStore.getCurrentTranscriptionStore().
						storeTranscription(t, null, null, nextPath, TranscriptionStore.EAF);
					}
				}
			} catch (IOException ioe) {
				System.out.println("Error while processing file: " + ioe.getMessage());
			}
			
			System.out.println("Processed file: " + file.getName());
		}
	}
	
	/**
	 * Replaces ECV's that have already been loaded. 
	 * 
	 * @param loadedMap a map of url's to lists of loaded ECV
	 * @param t the transcription to update
	 * @return false if not all ECV's have been replaced
	 */
	private boolean updateAlreadyLoadedCVS(HashMap<String, List<ExternalCV>> loadedMap, TranscriptionImpl t) {
		int numReplacedECV = 0;
		boolean allCVSInMap = true;
		for (int i = 0; i < t.getControlledVocabularies().size(); i++) {
			ControlledVocabulary cv = t.getControlledVocabularies().get(i);
			if (cv instanceof ExternalCV) {
				ExternalCV ecv = (ExternalCV) cv;
				if (!ecv.isLoadedFromURL()) {
					ExternalReference er = ecv.getExternalRef();
					if (er.getReferenceType() == ExternalReference.EXTERNAL_CV) {
						String urlString = er.getValue();
						if (urlString != null) {
							List<ExternalCV> loadedECV = loadedMap.get(urlString);
							if (loadedECV != null) {
								for (ExternalCV lECV : loadedECV) {
									if (lECV.getName().equals(ecv.getName())) {
										// already loaded, replace
										// t.replaceControlledVocabulary() changes the order of the CV's
										
										t.getControlledVocabularies().remove(i);
										t.getControlledVocabularies().add(i, lECV);
										numReplacedECV++;
										t.setUnchanged();
									}
								}
							} else {
								allCVSInMap = false;
							}
						}
					}
				}
			}
		}
		
		if (numReplacedECV > 0) {
			updateAnnotations(t);
		}
		
		return allCVSInMap;
	}
	
	/**
	 * The actual updating of the annotations. If no language is enforced it simply calls
	 * the existing checkAnnotECVConsistency() method. Otherwise it will loop over all tiers etc.
	 * 
	 * @param t the transcription to check
	 */
	private void updateAnnotations(TranscriptionImpl t) {
		if (forcedLanguage == null) {
			t.checkAnnotECVConsistency();
		} else {// alternative update of annotations
			// TODO the following code is partially copied from TranscriptionImpl. It would be possible to 
			// (temporarily) set the language attribute of the transcription and all tiers to the specified
			// language if this would surpass all preference settings (and if it doesn't set any preference)
			for (TierImpl tier : t.getTiers()) {
				LinguisticType lt = tier.getLinguisticType();
				String cvname = lt.getControlledVocabularyName();
				if (cvname == null || cvname.isEmpty()) {
					continue;
				}
				
				// Can we find that CV?
				ControlledVocabulary cv = t.getControlledVocabulary(cvname);
				if (cv == null) {
					continue;
				}
				
				if (cv instanceof ExternalCV) {// maybe a check on internal CV's would make sense as well?
					ExternalCV ecv = (ExternalCV) cv;
					
					if (!ecv.isLoadedFromURL() && !ecv.isLoadedFromCache()) {
						continue;
					}
					
					int langIndex = cv.getIndexOfLanguage(forcedLanguage);
					
					if (langIndex < 0) {
						continue;// not there
					}
					
					for (AbstractAnnotation currentAnn : tier.getAnnotations()) {
						String cvEntryRefId = currentAnn.getCVEntryId();
						if (cvEntryRefId == null) {
							// if there is no ref to a cv entry but there is an annotation value
							// check if the value is in the CV and add a reference to that entry
							if (!currentAnn.getValue().isEmpty()) {
								CVEntry cvEntry = ecv.getEntryWithValue(langIndex, currentAnn.getValue());
								if (cvEntry != null) {
									currentAnn.setCVEntryId(cvEntry.getId());
								}
							}
							continue;
						}
						
						CVEntry entry = ecv.getEntrybyId(cvEntryRefId);
						String value;
						if (entry != null && 
								(value = entry.getValue(langIndex)) != null &&
								!value.equals(currentAnn.getValue())) {
							currentAnn.setValue(value);
							t.setChanged();
						} else if (entry == null) {
							// The entry the External Ref. refers to is no longer
							// there, so remove the CV Entry ID from the annotation
							currentAnn.setCVEntryId(null);
							t.setChanged();
						}
					}
				}
			}
		}
	}
	
}
